from pygr.sqlgraph import *
from pygr.sequence import *
from pygr.seqdb import *


def buildClusterSpliceGraph(c, alt5, alt3):
    """use exon/splice start and end positions to build splice graph
    for a cluster c. Also finds exons that share same start (but differ
    at end: alt5), or share the same end (but differ at start: alt3)."""
    start = {}
    end = {}
    none = []
    for e in c.exons:
        if e.genomic_start not in start:
            start[e.genomic_start] = []
        start[e.genomic_start].append(e)
        if e.genomic_end not in end:
            end[e.genomic_end] = []
        end[e.genomic_end].append(e)
    for s in c.splices:
        try:
            exons1 = end[s.ver_gen_start]
        except KeyError:
            exons1 = none
        try:
            exons2 = start[s.ver_gen_end]
        except KeyError:
            exons2 = none
        for e1 in exons1:
            for e2 in exons2:
                e1.next[e2] = s # SAVE SPLICE AS EDGE INFO...
                s.exons = (e1, e2) # SAVE EXONS DIRECTLY ON THE SPLICE OBJECT
    for exons in start.values():
        for e1 in exons:
            for e2 in exons:
                if e1 != e2:
                    alt5 += e1
                    alt5 += e2
                    e1.alt5 += e2
                    e2.alt5 += e1
    for exons in end.values():
        for e1 in exons:
            for e2 in exons:
                if e1 != e2:
                    alt3 += e1
                    alt3 += e2
                    e1.alt3 += e2
                    e2.alt3 += e1


def loadCluster(c, exon_forms, splices, clusterExons, clusterSplices,
                spliceGraph, alt5, alt3):
    """Loads data for a single cluster, and builds it into a splice graph."""
    clusterExons += c
    clusterSplices += c
    for e in exon_forms.select('where cluster_id=%s', (c.id, )):
        c.exons += e
        spliceGraph += e
    for s in splices.select('where cluster_id=%s', (c.id, )):
        c.splices += s
    buildClusterSpliceGraph(c, alt5, alt3)


class ExonForm(TupleO, SeqPath): # ADD ATTRIBUTES STORING SCHEMA INFO

    def __init__(self, t):
        TupleO.__init__(self, t) # 1ST INITIALIZE ATTRIBUTE ACCESS
        SeqPath.__init__(self, g[self.cluster_id], # INITIALIZE AS SEQ INTERVAL
                         self.genomic_start - 1, self.genomic_end)

    def __getattr__(self, attr):
        'both parent classes have getattr, so have to call them both...'
        try:
            return TupleO.__getattr__(self, attr)
        except AttributeError:
            return SeqPath.__getattr__(self, attr)


class Splice(TupleO):
    pass


def loadSpliceGraph(jun03, cluster_t, exon_t, splice_t, genomic_seq_t,
                    mrna_seq_t=None, protein_seq_t=None, loadAll=True):
    """
    Build a splice graph from the specified SQL tables representing gene
    clusters, exon forms, and splices. Each table must be specified
    as a DB.TABLENAME string.
    These tables are loaded into memory.
    The splice graph is built based on exact match of exon and splice ends.
    In addition, also builds alt5Graph (exons that match at start but differ
    at end) and alt3Graph (exons that match at end but differ at start).

    Loads all cluster, exon and splice data if loadAll is True.

    Returns tuple: clusters, exons, splices, spliceGraph, alt5Graph, alt3Graph
    """

    # CREATE OUR GRAPHS
    clusterExons = dictGraph()
    clusterSplices = dictGraph()
    spliceGraph = dictGraph()
    alt5 = dictGraph()
    alt3 = dictGraph()

    class YiGenomicSequence(DNASQLSequence):

        def __len__(self):
            return self._select('length(seq)')  # USE SEQ LENGTH FROM DATABASE

    g = jun03[genomic_seq_t]
    # Force genomic seq table to use transparent access
    g.objclass(YiGenomicSequence)

    # Only process this if provided an mRNA table by the user.
    if mrna_seq_t is not None:
        mrna = jun03[mrna_seq_t]
        # Force mRNA seq table to use transparent access.
        mrna.objclass(SQLSequence)
    else:
        mrna = None

    # Only process this if provided a protein table by the user.
    if protein_seq_t is not None:

        class YiProteinSQLSequence(ProteinSQLSequence):

            def __len__(self):
                return self.protein_length # USE SEQ LENGTH FROM DATABASE

        protein = jun03[protein_seq_t]
        # Force protein seq table to use transparent access
        protein.objclass(YiProteinSQLSequence)
        # Alias 'protein_seq' to appear as 'seq'
        protein.addAttrAlias(seq='protein_seq')
    else:
        protein = None

    exon_forms = jun03[exon_t]
    ExonForm.__class_schema__ = SchemaDict(((spliceGraph, 'next'),
                                            (alt5, 'alt5'), (alt3, 'alt3')))
    # Bind this class to container as the one to use as "row objects".
    exon_forms.objclass(ExonForm)

    if loadAll:
        print 'Loading %s...' % exon_forms
        exon_forms.load(ExonForm)

    clusters = jun03[cluster_t]

    class Cluster(TupleO):
        __class_schema__ = SchemaDict(((clusterExons, 'exons'),
                                       (clusterSplices, 'splices')))
    # Bind this class to container as the one to use as "row objects".
    clusters.objclass(Cluster)
    if loadAll:
        print 'Loading %s...' % clusters
        clusters.load(Cluster)

    splices = jun03[splice_t]
    # Bind this class to container as the one to use as "row objects".
    splices.objclass(Splice)
    if loadAll:
        print 'Loading %s...' % splices
        splices.load(Splice)

##     print 'Saving alignment of protein to mrna isoforms...'
##     mrna_protein=PathMapping2()
##     for form_id in protein:
##         p=protein[form_id]
##         m=mrna[form_id]
##         start=3*(p.mRNA_start-1)+int(p.reading_frame)
##         end=start+3*p.protein_length
##         mrna_protein[p]=m[start:end]

        print 'Adding clusters to graph...'
        for c in clusters.values(): # ADD CLUSTERS AS NODES TO GRAPH
            clusterExons+=c
            clusterSplices+=c

        print 'Adding exons to graph...'
        for e in exon_forms.values():
            c=clusters[e.cluster_id]
            try:
                c.exons+=e
                spliceGraph+=e
            except IndexError:
                pass # BAD EXON: EMPTY SEQUENCE INTERVAL... IGNORE IT

        print 'Adding splices to graph...'
        for s in splices.values():
            try:
                c=clusters[s.cluster_id]
            except KeyError: # WIERD, ONE SPLICE WITH BLANK (NOT NULL) VALUE!
                pass
            else:
                c.splices+=s

        print 'Building splice graph...'
        for c in clusters.values():
            buildClusterSpliceGraph(c, alt5, alt3)

    return clusters, exon_forms, splices, g, spliceGraph, alt5, alt3, mrna,\
            protein, clusterExons, clusterSplices



from __future__ import generators
from sequence import *
import classutil
import UserDict
import weakref


def getAnnotationAttr(self, attr):
    'forward attributes from slice object if available'
    return self.db.getSliceAttr(self.db.sliceDB[self.id], attr)


def annotation_repr(self):
    if self.annotationType is not None:
        title = self.annotationType
    else:
        title = 'annot'
    if self.orientation > 0:
        return '%s%s[%d:%d]' % (title, self.id, self.start, self.stop)
    else:
        return '-%s%s[%d:%d]' % (title, self.id, -self.stop, -self.start)


class AnnotationSeqDescr(object):
    'get the sequence interval corresponding to this annotation'

    def __get__(self, obj, objtype):
        return absoluteSlice(obj._anno_seq, obj._anno_start,
                             obj._anno_start + obj.stop)


class AnnotationSliceDescr(object):
    'get the sequence interval corresponding to this annotation'

    def __get__(self, obj, objtype):
        return relativeSlice(obj.pathForward.sequence, obj.start, obj.stop)


class AnnotationSeqtypeDescr(object):
    'get seqtype of the sequence interval corresponding to this annotation'

    def __get__(self, obj, objtype):
        return obj._anno_seq.seqtype()


class AnnotationSeq(SeqPath):
    'base class representing an annotation'
    start = 0
    step = 1
    orientation = 1

    def __init__(self, id, db, parent, start, stop):
        self.id = id
        self.db = db
        self.stop = stop - start
        self._anno_seq = parent
        self._anno_start = start
        self.path = self

    __getattr__ = getAnnotationAttr
    sequence = AnnotationSeqDescr()
    annotationType = classutil.DBAttributeDescr('annotationType')
    _seqtype = AnnotationSeqtypeDescr()
    __repr__ = annotation_repr

    def __cmp__(self, other):
        if not isinstance(other, AnnotationSeq):
            return -1
        if cmp(self.sequence, other.sequence) == 0:
            if self.id == other.id and self.db is other.db:
                return cmp((self.start, self.stop), (other.start, other.stop))
        return NOT_ON_SAME_PATH

    def strslice(self, start, stop):
        raise ValueError('''this is an annotation, and you cannot get
                         a sequence string from it. Use its sequence attribute
                         to get a sequence object representing this interval.
                         ''')


class AnnotationSlice(SeqDBSlice):
    'represents subslice of an annotation'
    __getattr__=getAnnotationAttr
    sequence = AnnotationSliceDescr()
    annotationType = classutil.DBAttributeDescr('annotationType')
    __repr__ = annotation_repr


class TranslationAnnotSeqDescr(object):
    'get the sequence interval corresponding to this annotation'

    def __get__(self, obj, objtype):
        return absoluteSlice(obj._anno_seq, obj._anno_start, obj._anno_stop)


class TranslationAnnotFrameDescr(object):
    """Get the frame of this protein translation, relative to original DNA."""

    def __get__(self, obj, objtype):
        orig = obj.pathForward.sequence
        if orig.orientation > 0:
            frame = (orig.start % 3) + 1
        else:
            return -((orig.start + 1) % 3 + 1)
        return frame


class TranslationAnnot(AnnotationSeq):
    'annotation representing aa translation of a given nucleotide interval'

    def __init__(self, id, db, parent, start, stop):
        AnnotationSeq.__init__(self, id, db, parent, start, stop)
        self.stop /= 3
        self._anno_stop = stop
    sequence = TranslationAnnotSeqDescr()
    frame = TranslationAnnotFrameDescr()
    _seqtype = PROTEIN_SEQTYPE

    def strslice(self, start, stop):
        'get the aa translation of our associated ORF'
        try:
            aa = self._translation
        except AttributeError:
            aa = self._translation = translate_orf(str(self.sequence))
        return aa[start:stop]


class TranslationAnnotSliceDescr(object):
    'get the sequence interval corresponding to this annotation'

    def __get__(self, obj, objtype):
        return relativeSlice(obj.pathForward.sequence, 3 * obj.start,
                             3 * obj.stop)


class TranslationAnnotSlice(AnnotationSlice):
    sequence = TranslationAnnotSliceDescr()
    frame = TranslationAnnotFrameDescr()


class AnnotationDB(object, UserDict.DictMixin):
    'container of annotations as specific slices of db sequences'

    def __init__(self, sliceDB, seqDB, annotationType=None,
                 itemClass=AnnotationSeq,
                 itemSliceClass=AnnotationSlice,
                 itemAttrDict=None, # GET RID OF THIS BACKWARDS-COMPATIBILITY KLUGE!!
                 sliceAttrDict=None, maxCache=None, autoGC=True,
                 checkFirstID=True, **kwargs):
        '''sliceDB must map identifier to a sliceInfo object;
        sliceInfo must have attributes: id, start, stop, orientation;
        seqDB must map sequence ID to a sliceable sequence object;
        sliceAttrDict gives optional dict of item attributes that
        should be mapped to sliceDB item attributes.
        maxCache specfies the maximum number of annotation objects
        to keep in the cache.'''
        if autoGC: # automatically garbage collect unused objects
            self._weakValueDict = classutil.RecentValueDictionary(autoGC)
        else:
            self._weakValueDict = {} # object cache
        self.autoGC = autoGC
        if sliceAttrDict is None:
            sliceAttrDict = {}
        if sliceDB is not None:
            self.sliceDB = sliceDB
        else: # NEED TO CREATE / OPEN A DATABASE FOR THE USER
            self.sliceDB = classutil.get_shelve_or_dict(**kwargs)
        self.seqDB = seqDB
        self.annotationType = annotationType
        self.itemClass = itemClass
        self.itemSliceClass = itemSliceClass
        self.sliceAttrDict = sliceAttrDict # USER-PROVIDED ALIASES
        if maxCache is not None:
            self.maxCache = maxCache
        if checkFirstID:
            try: # don't cache anything now; schema may change itemClass!
                k = iter(self).next() # get the first ID if any
                self.get_annot_obj(k, self.sliceDB[k]) # valid annotation?
            except KeyError: # a convenient warning to the user...
                raise KeyError('''\
cannot create annotation object %s; sequence database %s may not be correct'''
                               % (k, repr(seqDB), ))
            except StopIteration:
                pass # dataset is empty so there is nothing we can check...
    __getstate__ = classutil.standard_getstate ############### PICKLING METHODS
    __setstate__ = classutil.standard_setstate
    _pickleAttrs = dict(sliceDB=0, seqDB=0, annotationType=0, autoGC=0,
                        itemClass=0, itemSliceClass=0, sliceAttrDict=0,
                        maxCache=0)

    def __hash__(self):                 # @CTB unnecessary??
        'ALLOW THIS OBJECT TO BE USED AS A KEY IN DICTS...'
        return id(self)

    def __getitem__(self, k):
        'get annotation object by its ID'
        try: # GET FROM OUR CACHE
            return self._weakValueDict[k]
        except KeyError:
            pass
        return self.sliceAnnotation(k, self.sliceDB[k])

    def __setitem__(self, k, v):
        raise KeyError('''you cannot save annotations directly using annoDB[k]
                       = v. Instead, use annoDB.new_annotation(k,sliceInfo)
                       where sliceInfo provides a sequence ID, start, stop (and
                       any additional info desired), and will be saved directly
                       to the sliceDB.''')

    def getSliceAttr(self, sliceInfo, attr):
        try:
            k = self.sliceAttrDict[attr] # USE ALIAS IF PROVIDED
        except KeyError:
            return getattr(sliceInfo, attr) # GET ATTRIBUTE AS USUAL
        try: # REMAP TO ANOTHER ATTRIBUTE NAME
            return getattr(sliceInfo, k)
        except TypeError: # TREAT AS int INDEX INTO A TUPLE
            return sliceInfo[k]

    def get_annot_obj(self, k, sliceInfo):
        'create an annotation object based on the input sliceInfo'
        start = int(self.getSliceAttr(sliceInfo, 'start'))
        stop = int(self.getSliceAttr(sliceInfo, 'stop'))

        try:
            orientation = self.getSliceAttr(sliceInfo, 'orientation')
            orientation = int(orientation)
            if orientation < 0 and start >= 0:
                start, stop = (-stop, -start) # Negative-orientation coords
        except (AttributeError, IndexError):
            pass                        # ok if no orientation is specified.

        if start >= stop:
            raise IndexError('annotation %s has zero or negative length \
                             [%s:%s]!' % (k, start, stop))
        seq_id = self.getSliceAttr(sliceInfo, 'id')
        seq = self.seqDB[seq_id]
        return self.itemClass(k, self, seq, start, stop)

    def sliceAnnotation(self, k, sliceInfo, limitCache=True):
        'create annotation and cache it'
        a = self.get_annot_obj(k, sliceInfo)
        try: # APPLY CACHE SIZE LIMIT IF ANY
            if limitCache and self.maxCache < len(self._weakValueDict):
                self._weakValueDict.clear()
        except AttributeError:
            pass
        self._weakValueDict[k] = a # CACHE THIS IN OUR DICT
        return a

    def new_annotation(self, k, sliceInfo):
        'save sliceInfo to the annotation database \
                and return annotation object'
        # First, check if it gives a valid annotation
        a = self.sliceAnnotation(k, sliceInfo)
        try:
            # Now, save it in the slice database
            self.sliceDB[k] = sliceInfo
        except:
            try:
                # Delete it from cache
                del self._weakValueDict[k]
            except:
                pass
            raise
        self._wroteSliceDB = True
        return a

    def foreignKey(self, attr, k):
        'iterate over items matching specified foreign key'
        for t in self.sliceDB.foreignKey(attr, k):
            try: # get from cache if exists
                yield self._weakValueDict[t.id]
            except KeyError:
                yield self.sliceAnnotation(t.id, t)

    def __contains__(self, k):
        return k in self.sliceDB

    def __len__(self):
        return len(self.sliceDB)

    def __iter__(self):
        return iter(self.sliceDB) ########## ITERATORS

    def  keys(self):
        return self.sliceDB.keys()

    def iteritems(self):
        'uses maxCache to manage caching of annotation objects'
        for k, sliceInfo in self.sliceDB.iteritems():
            yield k, self.sliceAnnotation(k, sliceInfo)

    def itervalues(self):
        'uses maxCache to manage caching of annotation objects'
        for k, v in self.iteritems():
            yield v

    def items(self):
        'forces load of all annotation objects into cache'
        return [(k, self.sliceAnnotation(k, sliceInfo, limitCache=False))
                for (k, sliceInfo) in self.sliceDB.items()]

    def values(self):
        'forces load of all annotation objects into cache'
        return [self.sliceAnnotation(k, sliceInfo, limitCache=False)
                for (k, sliceInfo) in self.sliceDB.items()]

    def add_homology(self, seq, search, id=None, idFormat='%s_%d',
                     autoIncrement=False, maxAnnot=999999,
                     maxLoss=None, sliceInfo=None, **kwargs):
        'find homology in our seq db and add as annotations'
        try: # ENSURE THAT sliceAttrDict COMPATIBLE WITH OUR TUPLE FORMAT
            if self.sliceAttrDict['id'] != 0:
                raise KeyError
        except KeyError:
            sliceAttrDict['id'] = 0 # USE TUPLE AS OUR INTERNAL STANDARD FORMAT
            sliceAttrDict['start'] = 1
            sliceAttrDict['stop'] = 2
        if autoIncrement:
            id = len(self.sliceDB)
        elif id is None:
            id = seq.id
        if isinstance(search, str): # GET SEARCH METHOD
            search = getattr(self.seqDB, search)
        if isinstance(seq, str): # CREATE A SEQ OBJECT
            seq = Sequence(seq, str(id))
        al = search(seq, **kwargs) # RUN THE HOMOLOGY SEARCH
        if maxLoss is not None: # REQUIRE HIT BE AT LEAST A CERTAIN LENGTH
            kwargs['minAlignSize'] = len(seq)-maxLoss
        hits = al[seq].keys(**kwargs) # OBTAIN LIST OF HIT INTERVALS
        if len(hits) > maxAnnot:
            raise ValueError('too many hits for %s: %d' % (id, len(hits)))
        out = []
        i = 0
        k = id
        for ival in hits: # CREATE ANNOTATION FOR EACH HIT
            if len(hits)>1: # NEED TO CREATE AN ID FOR EACH HIT
                if autoIncrement:
                    k = len(self.sliceDB)
                else:
                    k = idFormat % (id, i)
                i += 1
            if sliceInfo is not None: # SAVE SLICE AS TUPLE WITH INFO
                a = self.new_annotation(k, (ival.id, ival.start, ival.stop)
                                        + sliceInfo)
            else:
                a = self.new_annotation(k, (ival.id, ival.start, ival.stop))
            out.append(a) # RETURN THE ANNOTATION
        return out

    def close(self):
        'if sliceDB needs to be closed, do it and return True, otherwise False'
        try:
            if self._wroteSliceDB:
                self.sliceDB.close()
                self._wroteSliceDB = False # DISK FILE IS UP TO DATE
                return True
        except AttributeError:
            pass
        return False

    def __del__(self):
        if self.close():
            import sys
            print >>sys.stderr, '''
WARNING: you forgot to call AnnotationDB.close() after writing
new annotation data to it.  This could result in failure to properly
store the data in the associated disk file.  To avoid this, we
have automatically called AnnotationDB.sliceDB.close() to write the data
for you, when the AnnotationDB was deleted.'''

    def clear_cache(self):
        'empty the cache'
        self._weakValueDict.clear()

    # not clear what this should do for AnnotationDB

    def copy(self):
        raise NotImplementedError("nonsensical in AnnotationDB")

    def setdefault(self, k, d=None):
        raise NotImplementedError("nonsensical in AnnotationDB")

    def update(self, other):
        raise NotImplementedError("nonsensical in AnnotationDB")

    # these methods should not be implemented for read-only database.

    def clear(self):
        raise NotImplementedError("no deletions allowed")

    def pop(self):
        raise NotImplementedError("no deletions allowed")

    def popitem(self):
        raise NotImplementedError("no deletions allowed")


class AnnotationServer(AnnotationDB):
    'XMLRPC-ready server for AnnotationDB'
    xmlrpc_methods={'get_slice_tuple': 0, 'get_slice_items': 0,
                    'get_annotation_attr': 0, 'keys': 0,
                    '__len__': 0, '__contains__': 0}

    def get_slice_tuple(self, k):
        'get (seqID,start,stop) for a given key'
        try:
            sliceInfo = self.sliceDB[k]
        except KeyError:
            return '' # XMLRPC-acceptable failure code
        start = int(self.getSliceAttr(sliceInfo, 'start'))
        stop = int(self.getSliceAttr(sliceInfo, 'stop'))
        try:
            if int(self.getSliceAttr(sliceInfo, 'orientation')) < 0 \
               and start >= 0:
                start, stop = (-stop, -start) # Negative-orientation coords
        except AttributeError:
            pass
        return (self.getSliceAttr(sliceInfo, 'id'), start, stop)

    def get_slice_items(self):
        'get all (key,tuple) pairs in one query'
        l = []
        for k in self.sliceDB:
            l.append((k, self.get_slice_tuple(k)))
        return l

    def get_annotation_attr(self, k, attr):
        'get the requested attribute of the requested key'
        try:
            sliceInfo = self.sliceDB[k]
        except KeyError:
            return ''
        try:
            return self.getSliceAttr(sliceInfo, attr)
        except AttributeError:
            return ''


class AnnotationClientSliceDB(dict):
    'proxy just queries the server'

    def __init__(self, db):
        self.db = db
        dict.__init__(self)

    def __getitem__(self, k):
        try:
            return dict.__getitem__(self, k)
        except KeyError:
            t = self.db.server.get_slice_tuple(k)
            if t == '':
                raise KeyError('no such annotation: ' + str(k))
            dict.__setitem__(self, k, t)
            return t

    def __setitem__(self, k, v):
        raise ValueError('XMLRPC client is read-only')

    def keys(self):
        return self.db.server.keys()

    def __iter__(self):
        return iter(self.keys())

    def items(self):
        return self.db.server.get_slice_items()

    def iteritems(self):
        return iter(self.items())

    def __len__(self):
        return self.db.server.__len__()

    def __contains__(self, k):
        return self.db.server.__contains__(k)


class AnnotationClient(AnnotationDB):
    'XMLRPC AnnotationDB client'

    def __init__(self, url, name, seqDB, itemClass=AnnotationSeq,
                 itemSliceClass=AnnotationSlice, autoGC=True, **kwargs):
        if autoGC: # automatically garbage collect unused objects
            self._weakValueDict = classutil.RecentValueDictionary(autoGC)
        else:
            self._weakValueDict = {} # object cache
        self.autoGC = autoGC
        import coordinator
        self.server = coordinator.get_connection(url, name)
        self.url = url
        self.name = name
        self.seqDB = seqDB
        self.sliceDB = AnnotationClientSliceDB(self)
        self.itemClass = itemClass
        self.itemSliceClass = itemSliceClass

    def __getstate__(self):
        return dict(url=self.url, name=self.name, seqDB=self.seqDB,
                    autoGC=self.autoGC)

    def getSliceAttr(self, sliceInfo, attr):
        if attr=='id':
            return sliceInfo[0]
        elif attr=='start':
            return sliceInfo[1]
        elif attr=='stop':
            return sliceInfo[2]
        elif attr=='orientation':
            raise AttributeError('ori not saved')
        else:
            v = self.server.get_annotation_attr(sliceInfo[0], attr)
            if v=='':
                raise AttributeError('this annotation has no attr: ' + attr)
            return v




DNA_SEQTYPE=0
RNA_SEQTYPE=1
PROTEIN_SEQTYPE=2


def guess_seqtype(s):
    dna_letters='AaTtUuGgCcNn'
    ndna=0
    nU=0
    nT=0
    for l in s:
        if l in dna_letters:
            ndna += 1
        if l=='U' or l=='u':
            nU += 1
        elif l=='T' or l=='t':
            nT += 1
    ratio=ndna/float(len(s))
    if ratio>0.85:
        if nT>nU:
            return DNA_SEQTYPE
        else:
            return RNA_SEQTYPE
    else:
        return PROTEIN_SEQTYPE


seq_id_counter=0


def new_seq_id():
    global seq_id_counter
    seq_id_counter += 1
    return str(seq_id_counter-1)


def write_fasta(ofile, s, chunk=60, id=None, reformatter=None):
    "Trivial FASTA output"
    if id is None:
        try:
            id = str(s.id)
        except AttributeError:
            id = new_seq_id()

    ofile.write('>' + id + '\n')
    seq = str(s)
    if reformatter is not None: # APPLY THE DESIRED REFORMATTING
        seq = reformatter(seq)
    end = len(seq)
    pos = 0
    while 1:
        ofile.write(seq[pos:pos+chunk] + '\n')
        pos += chunk
        if pos >= end:
            break
    return id # IN CASE CALLER WANTS TEMP ID WE MAY HAVE ASSIGNED


def read_fasta(ifile):
    "iterate over id,title,seq from stream ifile"
    id = None
    isEmpty = True
    for line in ifile:
        if '>' == line[0]:
            if id is not None and len(seq) > 0:
                yield id, title, seq
                isEmpty = False
            id = line[1:].split()[0]
            title = line[len(id)+2:]
            seq = ''
        elif id is not None: # READ SEQUENCE
            for word in line.split(): # GET RID OF WHITESPACE
                seq += word
    if id is not None and len(seq) > 0:
        yield id, title, seq
    elif isEmpty:
        raise IOError('no readable sequence in FASTA file!')


def read_fasta_one_line(ifile): # @CTB deprecated; remove
    "read a single sequence line, return id,title,seq"
    id = None
    seq = ''
    while True:
        line = ifile.readline(1024) # READ AT MOST 1KB
        if line == '': # EOF
            break
        elif '>' == line[0]:
            id = line[1:].split()[0]
            title = line[len(id)+2:]
        elif id is not None: # READ SEQUENCE
            for word in line.split(): # GET RID OF WHITESPACE
                seq += word
            if len(seq) > 0:
                return id, title, seq
    raise IOError('no readable sequence in FASTA file!')


def read_fasta_lengths(ifile):
    "Generate sequence ID,length from stream ifile"
    id = None
    seqLength = 0
    isEmpty = True
    for line in ifile:
        if '>' == line[0]:
            if id is not None and seqLength > 0:
                yield id, seqLength
                isEmpty = False
            id = line[1:].split()[0]
            seqLength = 0
        elif id is not None: # READ SEQUENCE
            for word in line.split(): # GET RID OF WHITESPACE
                seqLength += len(word)
    if id is not None and seqLength > 0:
        yield id, seqLength
    elif isEmpty:
        raise IOError('no readable sequence in FASTA file!')


class AATranslation(object):
    'customizable translation class'
    geneticCode = dict(TTY='F', TTR='L', TCN='S', TAY='Y', TGY='C', TGG='W',
                       CTN='L', CCN='P', CAY='H', CAR='Q', CGN='R',
                       ATY='I', ATA='I', ATG='M', ACN='T', AAY='N', AAR='K',
                       AGY='S', AGR='R',
                       GTN='V', GCN='A', GAY='D', GAR='E', GGN='G',
                       TAR='*', TGA='*')

    def __init__(self):
        'initialize our translation dictionary by applying N,Y,R codes'
        geneticCode = self.geneticCode.copy()
        for codon, aa in self.geneticCode.items():
            if codon[2] == 'N':
                geneticCode[codon[:2]+'A'] = aa
                geneticCode[codon[:2]+'T'] = aa
                geneticCode[codon[:2]+'G'] = aa
                geneticCode[codon[:2]+'C'] = aa
            elif codon[2] == 'Y':
                geneticCode[codon[:2]+'T'] = aa
                geneticCode[codon[:2]+'C'] = aa
            elif codon[2] == 'R':
                geneticCode[codon[:2]+'A'] = aa
                geneticCode[codon[:2]+'G'] = aa
        self.geneticCode = geneticCode

    def __call__(self, s):
        'translate nucleotide string s to amino acid string'
        s = s.upper()
        s = s.replace('U', 'T')
        l = []
        for i in range(0, len(s), 3):
            try:
                l.append(self.geneticCode[s[i:i+3]])
            except KeyError:
                l.append('X') # uninterpretable
        return ''.join(l)

translate_orf = AATranslation() # default translation function





from __future__ import generators
from schema import *
import classutil


def update_graph(self, graph):
    'save nodes and edges of graph to self'
    for node, d in graph.iteritems():
        self += node
        saveDict = self[node]
        for target, edge in d.iteritems():
            saveDict[target] = edge


class PathList(list):
    """Internal representation for storing both nodes and edges as list
    So filter functions can see both nodes and edges"""

    def __init__(self, nodes=None, edges=None):
        if nodes != None:
            list.__init__(self, nodes)
        else:
            list.__init__(self)
        if edges != None:
            self.edge = list(edges)
        else:
            self.edge = []

    def append(self, val):
        list.append(self, val)
        self.edge.append(val)

    def extend(self, l):
        list.extend(self, l) # EXTEND TOP-LEVEL LIST AS USUAL
        try: # EXTEND OUR EDGE LIST AS WELL
            self.edge.extend(l.edge)
        except AttributeError: #IF l HAS NO EDGES, PAD OUR EDGE LIST WITH Nones
            self.edge.extend(len(l) * [None])


class Edge(list):
    "Interface to edge information."
    isDirected = False

    def __init__(self, graph, nodes, edgeInfo):
        self.graph = graph
        if edgeInfo:
            self.edgeInfo = edgeInfo
        list.__init__(self, nodes) # SAVE NODES AS TUPLE

    def __getattr__(self, attr):
        try:
            return getattr(self.edgeInfo, attr)
        except AttributeError:
            if isinstance(self.edgeInfo, types.DictType):
                # Treat edgeInfo as an attribute dictionary.
                return self.edgeInfo[attr]
            raise AttributeError(attr)

    # Should we define setattr here too, to allow users to add new attribute
    # values? The problem is setattr is painful to implement due to the
    # recursive reference problem.

    def __cmp__(self, other): # DO WE NEED TO COMPARE EDGE INFO??
        if not isinstance(other, Edge): # CAN ONLY COMPARE A PAIR OF EDGES
            return -1
        diff = cmp(self.graph, other.graph)
        if diff: # NOT IN THE SAME GRAPH...
            return diff
        elif self.isDirected: # IF DIRECTED, JUST COMPARE IN CURRENT ORDER
            return tuple.__cmp__(self, other)
        else: # UNDIRECTED COMPARISON REQUIRES PUTTING BOTH IN SAME ORDER
            me = [i for i in self]
            you = [i for i in other]
            me.sort()
            you.sort()
            return cmp(me, you)

    # NEEDS SCHEMA SUPPORT: RETURN A SINGLE SCHEMA TUPLE DESCRIBING THIS EDGE.

class DirectedEdge(Edge):
    isDirected = True


# need a class to provide access to the edges in a graph
# iterator, membership test
#class EdgeSet


class dictEdge(dict):
    """2nd layer graph interface implemenation using dict.
    """
    dictClass = dict

    def __init__(self, graph, fromNode):
        self.graph = graph
        self.fromNode = fromNode
        self.dictClass.__init__(self) # INITIALIZE TOPLEVEL DICTIONARY

    def __iadd__(self, target):
        "Add edge from fromNode to target with no edge-info"
        self[target] = None
        return self # THIS IS REQUIRED FROM iadd()!!

    def __setitem__(self, target, edgeInfo):
        "Add edge from fromNode to target with edgeInfo"
        self.dictClass.__setitem__(self, target, edgeInfo)
        if target not in self.graph: # ADD NEW NODE TO THE NODE DICT
            self.graph += target

    _setitem_ = dict.__setitem__ # INTERNAL INTERFACE FOR SAVING AN ENTRY

    def __delitem__(self, target):
        "Delete edge from fromNode to target"
        try:
            self.dictClass.__delitem__(self, target)
        except KeyError: # GENERATE A MORE INFORMATIVE ERROR MESSAGE
            raise KeyError('No edge from node to target')

    def __isub__(self, target):
        "Delete edge from fromNode to target"
        self.__delitem__(target)
        return self # THIS IS REQUIRED FROM iadd()!!

    def edges(self):
        "Return iterator for accessing edges from fromNode"
        for target, edgeInfo in self.items():
            if isinstance(edgeInfo, Edge):
                yield edgeInfo
            else:
                yield Edge(self.graph, (self.fromNode, target, edgeInfo),
                           edgeInfo)


class dictGraph(dict):
    """Top layer graph interface implemenation using dict.
    """
    dictClass = dict
    edgeDictClass = dictEdge

    def __init__(self, schema=None, domain=None, range=None):
        if schema and domain and range:
            if domain not in schema:
                schema += domain #ADD DOMAIN AS NODE TO schema GRAPH
            schema[domain][range] = self
        self.dictClass.__init__(self) # INITIALIZE TOPLEVEL DICTIONARY

    def __iadd__(self, node, ruleSet=False):
        "Add node to graph with no edges"
        if node not in self:
            self.dictClass.__setitem__(self, node, self.edgeDictClass(self,
                                                                      node))
            if ruleSet == False:
                ruleSet = getschema(node, graph=self)
            for rule in ruleSet:
                if isinstance(rule[1], types.StringType):
                    # Attribute binding; bind directly to attribute.
                    setattr(node, rule[1], self[node])
        return self # THIS IS REQUIRED FROM iadd()!!

    def __setitem__(self, node, target):
        "This method exists only to support g[n]+=o.  Do not use as g[n]=foo."
        if self[node] != target:
            raise ValueError('Incorrect usage. Add edges using g[n]+=o \
                             or g[n][o]=edge.')

    def __delitem__(self, node):
        "Delete node from graph."
        # Grr, we really need to find all edges going to this node
        # and delete them.
        try:
            # Do stuff to remove it here...
            self.dictClass.__delitem__(self, node)
        except KeyError:
            raise KeyError('Node not present in mapping.')
        for rule in getschema(node, graph=self):
            if isinstance(rule[1], types.StringType): # ATTRIBUTE BINDING!
                delattr(node, rule[1])  # REMOVE ATTRIBUTE BINDING

    def __isub__(self, node):
        "Delete node from graph"
        self.__delitem__(node)
        return self # THIS IS REQUIRED FROM isub()!!

    def __hash__(self): # SO SCHEMA CAN INDEX ON GRAPHS...
        return id(self)

    def edges(self):
        "Return iterator for all edges in this graph"
        for edgedict in self.values():
            for edge in edgedict.edges():
                yield edge
    update = update_graph


class dictEdgeFB(dictEdge):
    "dictEdge subclass that saves both forward and backward edges"

    def __setitem__(self, target, edgeInfo):
        "Save edge in both forward and backward dicts."
        dictEdge.__setitem__(self, target, edgeInfo) # FORWARD EDGE
        try:
            d = self.graph._inverse[target]
        except KeyError:
            d = self.dictClass()
            self.graph._inverse[target] = d
        d[self.fromNode] = edgeInfo # SAVE BACKWARD EDGE

    def __invert__(self):
        "Get nodes with edges to this node"
        return self.graph._inverse[self.fromNode]


class dictGraphFB(dictGraph):
    "Graph that saves both forward and backward edges"

    def __init__(self, **kwargs):
        dictGraph.__init__(self, **kwargs)
        self._inverse = self.dictClass()
    __invert__ = classutil.standard_invert

    def __delitem__(self, node):
        "Delete node from the graph"
        try:
            fromNodes = self._inverse[node]
            del self._inverse[node] # REMOVE FROM _inverse DICT
        except KeyError:
            pass
        else: # DELETE EDGES TO THIS NODE
            for i in fromNodes:
                del self[i][node]
        dictGraph.__delitem__(self, node)


def listUnion(ivals):
    'merge all items using union operator'
    union = None
    for ival in ivals:
        try:
            union += ival
        except TypeError:
            union = ival
    return union


class DictQueue(dict):
    'each index entry acts like a queue; setitem PUSHES, and delitem POPS'

    def __setitem__(self, k, val):
        try:
            dict.__getitem__(self, k).append(val)
        except KeyError:
            dict.__setitem__(self, k, [val])

    def __getitem__(self, k):
        return dict.__getitem__(self, k)[0]

    def __delitem__(self, k):
        l=dict.__getitem__(self, k)
        del l[0]
        if len(l) == 0:
            dict.__delitem__(self, k)


################################ PYGR.DATA.SCHEMA - AWARE CLASSES BELOW


def close_if_possible(self):
    'close storage to ensure any pending data is written'
    try:
        do_close = self.d.close
    except AttributeError:
        pass
    else:
        do_close()


class Collection(object):
    'flexible storage mapping ID --> OBJECT'

    def __init__(self, saveDict=None, dictClass=dict, **kwargs):
        '''saveDict, if not None, the internal mapping to use as our storage.
        filename: if provided, a file path to a shelve (BerkeleyDB) file to
              store the data in.
        dictClass: if provided, the class to use for storage of dict data.'''
        if saveDict is not None:
            self.d = saveDict
        elif 'filename' in kwargs: # USE A SHELVE (BERKELEY DB)
            try:
                if kwargs['intKeys']: # ALLOW INT KEYS, HAVE TO USE IntShelve
                    self.__class__ = IntShelve
                else:
                    raise KeyError
            except KeyError:
                self.__class__ = PicklableShelve
            return self.__init__(**kwargs)
        else:
            self.d = dictClass()
        classutil.apply_itemclass(self, kwargs)

    def __getitem__(self, k):
        return self.d[k]

    def __setitem__(self, k, v):
        self.d[k] = v

    def __delitem__(self, k):
        del self.d[k]

    def __len__(self):
        return len(self.d)

    def __contains__(self, k):
        return k in self.d

    def __iter__(self):
        return iter(self.d)

    def __getattr__(self, attr):
        if attr == '__setstate__' or attr == '__dict__':
            # This prevents infinite recursion during unpickling.
            raise AttributeError
        try: # PROTECT AGAINST INFINITE RECURSE IF NOT FULLY __init__ED...
            return getattr(self.__dict__['d'], attr)
        except KeyError:
            raise AttributeError('Collection has no subdictionary')
    close = close_if_possible

    def __del__(self):
        'must ensure that shelve object is closed to save pending data'
        try:
            self.close()
        except classutil.FileAlreadyClosedError:
            pass


class PicklableShelve(Collection):
    'persistent storage mapping ID --> OBJECT'

    def __init__(self, filename, mode=None, writeback=False,
                 unpicklingMode=False, verbose=True, **kwargs):
        '''Wrapper for a shelve object that can be pickled.  Ideally, you
should specify a TWO letter mode string: the first letter to
indicate what mode the shelve should be initially opened in, and
the second to indicate the mode to open the shelve during unpickling.
e.g. mode='nr': to create an empty shelve (writable),
which in future will be re-opened read-only.
Also, mode=None makes it first attempt to open read-only, but if the file
does not exist will create it using mode 'c'. '''
        # Mark this string as a file path.
        self.filename = classutil.SourceFileName(str(filename))
        self.writeback = writeback
        if unpicklingMode or mode is None or mode == 'r':
            # Just use mode as given.
            self.mode = mode
        elif mode == 'n' or mode == 'c' or mode == 'w':
            # Ambiguous modes, warn & set default.
            if verbose:
                import sys
                print >>sys.stderr, '''Warning: you opened shelve file %s
in mode '%s' but this is ambiguous for how the shelve should be
re-opened later during unpickling.  By default it will be
re-opened in mode 'r' (read-only).  To make it be re-opened
writable, create it in mode '%sw', or call its method
reopen('w'), which will make it be re-opened in mode 'w' now and
in later unpickling.  To suppress this warning message, use the
verbose=False option.''' % (filename, mode, mode)
            self.mode = 'r'
        else: # PROCESS UNAMBIGUOUS TWO-LETTER mode STRING
            try:
                if len(mode) == 2 and mode[0] in 'ncwr' and mode[1] in 'cwr':
                    self.mode = mode[1] # IN FUTURE OPEN IN THIS MODE
                    mode = mode[0] # OPEN NOW IN THIS MODE
                else:
                    raise ValueError('invalid mode string: ' + mode)
            except TypeError:
                raise ValueError('file mode must be a string!')
        if unpicklingMode:
            self.d = classutil.open_shelve(filename, mode, writeback,
                                           allowReadOnly=True)
        else:
            self.d = classutil.open_shelve(filename, mode, writeback)
        classutil.apply_itemclass(self, kwargs)

    __getstate__ = classutil.standard_getstate ############### PICKLING METHODS
    __setstate__ = classutil.standard_setstate
    _pickleAttrs = dict(filename=0, mode=0, writeback=0)

    def close(self):
        '''close our shelve index file.'''
        self.d.close()

    def __setitem__(self, k, v):
        try:
            self.d[k] = v
        except TypeError:
            raise TypeError('to allow int keys, you must pass intKeys=True \
to constructor!')

    def reopen(self, mode='r'):
        're-open shelve in the specified mode, and save mode on self'
        self.close()
        self.d = classutil.open_shelve(self.filename, mode,
                                       writeback=self.writeback)
        self.mode = mode


class IntShelve(PicklableShelve):
    'provides an interface to shelve that can use int as key'

    def saveKey(self, i):
        'convert to string key'
        if isinstance(i, int):
            return 'int:%s' % i
        elif isinstance(i, str):
            return i
        try:
            return 'int:%s' % int(i)
        except TypeError:
            pass
        raise KeyError('IntShelve can only save int or str as key')

    def trueKey(self, k):
        "convert back to key's original format"
        if k.startswith('int:'):
            return int(k[4:])
        else:
            return k

    def __getitem__(self, k):
        return self.d[self.saveKey(k)]

    def __setitem__(self, k, v):
        self.d[self.saveKey(k)] = v

    def __delitem__(self, k):
        del self.d[self.saveKey(k)]

    def __contains__(self, k):
        return self.saveKey(k) in self.d

    def __iter__(self): ################ STANDARD ITERATOR METHODS
        for k in self.d:
            yield self.trueKey(k)

    def keys(self):
        return [k for k in self]

    def iteritems(self):
        for k, v in self.d.iteritems():
            yield self.trueKey(k), v

    def items(self):
        return [k for k in self.iteritems()]


## PACKING / UNPACKING METHODS FOR SEPARATING INTERNAL VS. EXTERNAL
## REPRESENTATIONS OF GRAPH NODES AND EDGES
## 1. ID-BASED PACKING: USE obj.id AS INTERNAL REPRESENTATION
##
## 2. TRIVIAL: INTERNAL AND EXTERNAL REPRESENTATIONS IDENTICAL
##    WORKS WELL FOR STRING OR INT NODES / EDGES.
##
## 3. PICKLE PACKING: USE PICKLE AS INTERNAL REPRESENTATION
def pack_id(self, obj):
    'extract id attribute from obj'
    try:
        return obj.id
    except AttributeError:
        if obj is None:
            return None
        raise


def unpack_source(self, objID):
    return self.sourceDB[objID]


def unpack_target(self, objID):
    return self.targetDB[objID]


def unpack_edge(self, objID):
    try:
        return self.edgeDB[objID]
    except KeyError:
        if objID is None:
            return None
        raise


def add_standard_packing_methods(localDict):
    localDict['pack_source'] = pack_id
    localDict['pack_target'] = pack_id
    localDict['pack_edge'] = pack_id
    localDict['unpack_source'] = unpack_source
    localDict['unpack_target'] = unpack_target
    localDict['unpack_edge'] = unpack_edge


def add_trivial_packing_methods(localDict):
    for name in ('pack_source', 'pack_target', 'pack_edge',
                 'unpack_source', 'unpack_target', 'unpack_edge'):
        localDict[name] = lambda self, obj: obj


def pack_pickle(self, obj):
    'get pickle string for obj'
    import pickle
    return pickle.dumps(obj)


def unpack_pickle(self, s):
    'unpickle string to get obj'
    import pickle
    return pickle.loads(s)


class MappingInverse(object):

    def __init__(self, db):
        self._inverse = db
        self.attr = db.inverseAttr

    def __getitem__(self, k):
        return self._inverse.sourceDB[getattr(k, self.attr)]
    __invert__ = classutil.standard_invert


class Mapping(object):
    '''dict-like class suitable for persistent usages.  Extracts ID values from
    keys and values passed to it, and saves IDs into its internal dictionary
    instead of the actual objects.  Thus, the external interface is objects,
    but the internal storage is ID values.'''

    def __init__(self, sourceDB, targetDB, saveDict=None, IDAttr='id',
                 targetIDAttr='id', itemAttr=None, multiValue=False,
                 inverseAttr=None, **kwargs):
        '''sourceDB: dictionary that maps key ID values to key objects
        targetDB: dictionary that maps value IDs to value objects
        saveDict, if not None, is the internal mapping to use as our storage
        IDAttr: attribute name to obtain an ID from a key object
        targetIDAttr: attribute name to obtain an ID from a value object
        itemAttr, if not None, the attribute to obtain target (value) ID
           from an internal storage value
        multiValue: if True, treat each value as a list of values.
        filename: if provided, is a file path to a shelve (BerkeleyDB) file to
              store the data in.
        dictClass: if not None, is the class to use for storage of dict data'''
        if saveDict is None:
            self.d = classutil.get_shelve_or_dict(**kwargs)
        else:
            self.d = saveDict
        self.IDAttr = IDAttr
        self.targetIDAttr = targetIDAttr
        self.itemAttr = itemAttr
        self.multiValue = multiValue
        self.sourceDB = sourceDB
        self.targetDB = targetDB
        if inverseAttr is not None:
            self.inverseAttr = inverseAttr

    def __getitem__(self, k):
        kID = getattr(k, self.IDAttr)
        return self.getTarget(self.d[kID])

    def getTarget(self, vID):
        if self.itemAttr is not None:
            vID = getattr(vID, self.itemAttr)
        if self.multiValue:
            return [self.targetDB[j] for j in vID]
        else:
            return self.targetDB[vID]

    def __setitem__(self, k, v):
        if self.multiValue:
            v = [getattr(x, self.targetIDAttr) for x in v]
        else:
            v = getattr(v, self.targetIDAttr)
        self.d[getattr(k, self.IDAttr)] = v

    def __delitem__(self, k):
        del self.d[getattr(k, self.IDAttr)]

    def __contains__(self, k):
        return getattr(k, self.IDAttr) in self.d

    def __len__(self):
        return len(self.d)

    def clear(self):
        self.d.clear()

    def copy(self):
        return Mapping(self.sourceDB, self.targetDB, self.d.copy(),
                       self.IDAttr, self.targetIDAttr, self.itemAttr,
                       self.multiValue)

    def update(self, b):
        for k, v in b.iteritems():
            self[k] = v

    def get(self, k, v=None):
        try:
            return self[k]
        except KeyError:
            return v

    def setdefault(self, k, v=None):
        try:
            return self[k]
        except KeyError:
            self[k] = v
            return v

    def pop(self, k, v=None):
        try:
            v = self[k]
        except KeyError:
            return v
        del self[k]
        return v

    def popitem(self):
        kID, vID = self.d.popitem()
        return kID, self.getTarget(vID)

    def __iter__(self): ######################## ITERATORS
        for kID in self.d:
            yield self.sourceDB[kID]

    def keys(self):
        return [k for k in self]

    def itervalues(self):
        for vID in self.d.itervalues():
            yield self.getTarget(vID)

    def values(self):
        return [v for v in self.itervalues()]

    def iteritems(self):
        for kID, vID in self.d.iteritems():
            yield self.sourceDB[kID], self.getTarget(vID)

    def items(self):
        return [x for x in self.iteritems()]

    __invert__ = classutil.standard_invert
    _inverseClass = MappingInverse
    close = close_if_possible

    def __del__(self):
        close_if_possible(self)


def graph_cmp(self, other):
    'compare two graph dictionaries'
    import sys
    diff = cmp(len(self), len(other))
    if diff != 0:
        print >>sys.stderr, 'len diff:', len(self), len(other)
        return diff
    for node, d in self.iteritems():
        try:
            d2 = other[node]
        except KeyError:
            print >>sys.stderr, 'other missing key'
            return 1
        diff = cmp(d, d2)
        if diff != 0:
            print >>sys.stderr, 'value diff', d, d2
            return diff
    return 0


class IDNodeDict(object):
    """2nd layer graph interface implementation using proxy dict.
       e.g. shelve."""
    dictClass = dict

    def __init__(self, graph, fromNode):
        self.graph = graph
        self.fromNode = fromNode

    def __getitem__(self, target): ############# ACCESS METHODS
        edgeID = self.graph.d[self.fromNode][self.graph.pack_target(target)]
        return self.graph.unpack_edge(edgeID)

    def __setitem__(self, target, edgeInfo):
        "Add edge from fromNode to target with edgeInfo"
        self.graph.d[self.fromNode][self.graph.pack_target(target)] \
             = self.graph.pack_edge(edgeInfo)
        if not hasattr(self.graph, 'sourceDB') or \
           (hasattr(self.graph, 'targetDB') and \
           self.graph.sourceDB == self.graph.targetDB):
            self.graph += target # ADD NEW NODE TO THE NODE DICT

    def __delitem__(self, target):
        "Delete edge from fromNode to target"
        try:
            del self.graph.d[self.fromNode][self.graph.pack_target(target)]
        except KeyError: # GENERATE A MORE INFORMATIVE ERROR MESSAGE
            raise KeyError('No edge from node to target')

    ######### CONVENIENCE METHODS THAT USE THE ACCESS METHODS ABOVE
    def __iadd__(self, target):
        "Add edge from fromNode to target with no edge-info"
        self[target] = None
        return self # THIS IS REQUIRED FROM iadd()!!

    def __isub__(self, target):
        "Delete edge from fromNode to target"
        self.__delitem__(target)
        return self # THIS IS REQUIRED FROM iadd()!!

    def edges(self):
        "Return iterator for accessing edges from fromNode"
        for target, edgeInfo in self.graph.d[self.fromNode].items():
            yield (self.graph.unpack_source(self.fromNode),
                   self.graph.unpack_target(target),
                   self.graph.unpack_edge(edgeInfo))

    def __len__(self):
        return len(self.graph.d[self.fromNode])

    def keys(self):
        return [k[1] for k in self.edges()] ##### ITERATORS

    def values(self):
        return [k[2] for k in self.edges()]

    def items(self):
        return [k[1:3] for k in self.edges()]

    def __iter__(self):
        for source, target, edgeInfo in self.edges():
            yield target

    def itervalues(self):
        for source, target, edgeInfo in self.edges():
            yield edgeInfo

    def iteritems(self):
        for source, target, edgeInfo in self.edges():
            yield target, edgeInfo

    __cmp__ = graph_cmp


class IDNodeDictWriteback(IDNodeDict):
    'forces writing of subdictionary in shelve opened without writeback=True'

    def __setitem__(self, target, edgeInfo):
        d = self.graph.d[self.fromNode]
        d[self.graph.pack_target(target)] = self.graph.pack_edge(edgeInfo)
        self.graph.d[self.fromNode] = d # WRITE IT BACK... REQUIRED FOR SHELVE
        self.graph += target # ADD NEW NODE TO THE NODE DICT

    def __delitem__(self, target):
        d = self.graph.d[self.fromNode]
        del d[self.graph.pack_target(target)]
        self.graph.d[self.fromNode] = d # WRITE IT BACK... REQUIRED FOR SHELVE


class IDNodeDictWriteNow(IDNodeDictWriteback):
    'opens shelve for writing, writes an item, immediately reopens'

    def __setitem__(self, target, edgeInfo):
        self.graph.d.reopen('w')
        IDNodeDictWriteback.__setitem__(self, target, edgeInfo)
        self.graph.d.reopen('w')

    def __delitem__(self, target):
        self.graph.d.reopen('w')
        IDNodeDictWriteback.__delitem__(self, target)
        self.graph.d.reopen('w')


class IDGraphEdges(object):
    '''provides iterator over edges as (source, target, edge) tuples
       and getitem[edge] --> [(source, target), ...]'''

    def __init__(self, g):
        self.g = g

    def __iter__(self):
        for d in self.g.itervalues():
            for edge in d.edges():
                yield edge

    def __getitem__(self, edge):
        l = []
        for sourceID, targetID in self.d[edge.id]:
            l.append((self.g.sourceDB[sourceID], self.g.targetDB[targetID]))
        return l

    def __call__(self):
        return self


class IDGraphEdgeDescriptor(object):
    'provides interface to edges on demand'

    def __get__(self, obj, objtype):
        return IDGraphEdges(obj)


def save_graph_db_refs(self, sourceDB=None, targetDB=None, edgeDB=None,
                       simpleKeys=False, unpack_edge=None,
                       edgeDictClass=None, graph=None, **kwargs):
    'apply kwargs to reference DB objects for this graph'
    if sourceDB is not None:
        self.sourceDB = sourceDB
    else:
        # No source DB, store keys as internal representation.
        simpleKeys = True
    if targetDB is not None:
        self.targetDB=targetDB
    if edgeDB is not None:
        self.edgeDB=edgeDB
    else: # just save the edge object as itself (not its ID)
        self.pack_edge = self.unpack_edge = lambda edge: edge
    if simpleKeys: # SWITCH TO USING TRIVIAL PACKING: OBJECT IS ITS OWN ID
        self.__class__ = self._IDGraphClass
    if unpack_edge is not None:
        self.unpack_edge = unpack_edge # UNPACKING METHOD OVERRIDES DEFAULT
    if graph is not None:
        self.graph = graph
    if edgeDictClass is not None:
        self.edgeDictClass = edgeDictClass


def graph_db_inverse_refs(self, edgeIndex=False):
    'return kwargs for inverse of this graph, or edge index of this graph'
    if edgeIndex: # TO CONSTRUCT AN EDGE INDEX
        db = ('edgeDB', 'sourceDB', 'targetDB')
    else: # DEFAULT: TO CONSTRUCT AN INVERSE MAPPING OF THE GRAPH
        db = ('targetDB', 'sourceDB', 'edgeDB')
    try:
        d = dict(sourceDB=getattr(self, db[0]), targetDB=getattr(self, db[1]))
        try:
            d['edgeDB'] = getattr(self, db[2]) # EDGE INFO IS OPTIONAL
        except AttributeError:
            pass
    except AttributeError:
        d = dict(simpleKeys=True) # NO SOURCE / TARGET DB, SO USE IDs AS KEYS
    try: # COPY THE LOCAL UNPACKING METHOD, IF ANY
        if not edgeIndex:
            d['unpack_edge'] = self.__dict__['unpack_edge']
    except KeyError:
        pass
    return d


def graph_setitem(self, node, target):
    "This method exists only to support g[n]+=o.  Do not use as g[n]=foo."
    node = self.pack_source(node)
    try:
        if node == target.fromNode:
            return
    except AttributeError:
        pass
    raise ValueError('Incorrect usage. Add edges using g[n]+=o or \
g[n][o]=edge.')


class Graph(object):
    """Top layer graph interface implemenation using proxy dict.
       Works with dict, shelve, any mapping interface."""
    edgeDictClass = IDNodeDict # DEFAULT EDGE DICT

    def __init__(self, saveDict=None, dictClass=dict, writeNow=False,
                 **kwargs):
        if saveDict is not None: # USE THE SUPPLIED STORAGE
            self.d = saveDict
        elif 'filename' in kwargs: # USE A SHELVE (BERKELEY DB)
            try:
                if kwargs['intKeys']: # ALLOW INT KEYS, HAVE TO USE IntShelve
                    self.d = IntShelve(writeback=False, **kwargs)
                else:
                    raise KeyError
            except KeyError:
                self.d = PicklableShelve(writeback=False, **kwargs)
            if writeNow:
                # Write immediately.
                self.edgeDictClass = IDNodeDictWriteNow
            else:
                # Use our own writeback.
                self.edgeDictClass = IDNodeDictWriteback
        else:
            self.d = dictClass()
        save_graph_db_refs(self, **kwargs)

    __getstate__ = classutil.standard_getstate ############### PICKLING METHODS
    __setstate__ = classutil.standard_setstate
    _pickleAttrs = dict(d='saveDict', sourceDB=0, targetDB=0, edgeDB=0,
                        edgeDictClass=0)
    add_standard_packing_methods(locals())  ############ PACK / UNPACK METHODS

    def close(self):
        '''If possible, close our dict.'''
        try:
            do_close = self.d.close
        except AttributeError:
            pass
        else:
            do_close()

    def __len__(self):
        return len(self.d)

    def __iter__(self):
        for node in self.d:
            yield self.unpack_source(node)

    def keys(self):
        return [k for k in self]

    def itervalues(self):
        for node in self.d:
            yield self.edgeDictClass(self, node)

    def values(self):
        return [v for v in self.itervalues()]

    def iteritems(self):
        for node in self.d:
            yield self.unpack_source(node), self.edgeDictClass(self, node)

    def items(self):
        return [v for v in self.iteritems()]

    edges = IDGraphEdgeDescriptor()

    def __iadd__(self, node):
        "Add node to graph with no edges"
        node = self.pack_source(node)
        if node not in self.d:
            self.d[node] = {} # INITIALIZE TOPLEVEL DICTIONARY
        return self # THIS IS REQUIRED FROM iadd()!!

    def __contains__(self, node):
        return self.pack_source(node) in self.d

    def __getitem__(self, node):
        if node in self:
            return self.edgeDictClass(self, self.pack_source(node))
        raise KeyError('node not in graph')
    __setitem__ = graph_setitem

    def __delitem__(self, node):
        "Delete node from graph."
        node = self.pack_source(node)
        # Grr, we really need to find all edges going to this node
        # and delete them.
        try:
            del self.d[node]  # DO STUFF TO REMOVE IT HERE...
        except KeyError:
            raise KeyError('Node not present in mapping.')

    def __isub__(self, node):
        "Delete node from graph"
        self.__delitem__(node)
        return self # THIS IS REQUIRED FROM isub()!!
    update = update_graph
    __cmp__ = graph_cmp

    def __del__(self):
        try:
            self.close()
        except classutil.FileAlreadyClosedError:
            pass

# NEED TO PROVIDE A REAL INVERT METHOD!!
##     def __invert__(self):
##         'get an interface to the inverse graph mapping'
##         try: # CACHED
##             return self._inverse
##         except AttributeError: # NEED TO CONSTRUCT INVERSE MAPPING
##             self._inverse = IDGraph(~(self.d), self.targetDB, self.sourceDB,
##                                     self.edgeDB)
##             self._inverse._inverse = self
##             return self._inverse
##
    def __hash__(self): # SO SCHEMA CAN INDEX ON GRAPHS...
        return id(self)


class IDGraph(Graph):
    add_trivial_packing_methods(locals())

Graph._IDGraphClass = IDGraph


class KeepUniqueDict(dict):
    'dict that blocks attempts to overwrite an existing key'

    def __setitem__(self, k, v):
        try:
            if self[k] is v:
                return # ALREADY SAVED.  NOTHING TO DO!
        except KeyError: # NOT PRESENT, SO JUST SAVE THE VALUE
            dict.__setitem__(self, k, v)
            return
        raise KeyError('attempt to overwrite existing key!')

    def __hash__(self):
        'ALLOW THIS OBJECT TO BE USED AS A KEY IN DICTS...'
        return id(self)





def catalog_downloads(url, fileFilter, fileNamer, fileDocumenter, klass):
    '''returns dict of resources for downloading target data:
    url: URL to a directory that gives links to the desired data
    fileFilter: function that returns True for a desired file name
    fileNamer: function that converts a file name into desired key value
    fileDocumenter: function that returns a docstring based on file name
    klass: a Python class or function that takes file URL as a single argument,
           and returns value to store in dictionary

    example usage:
    d = catalog_downloads("http://biodb.bioinformatics.ucla.edu/PYGRDATA/",
                          lambda x: x.endswith(".txt.gz"),
                          lambda x: "Bio.MSA.UCSC."+x[:-3],
                          SourceURL)
    '''
    import formatter
    import htmllib
    import urllib
    #from BeautifulSoup import BeautifulSoup
    ifile = urllib.urlopen(url)
    try:
        p = htmllib.HTMLParser(formatter.NullFormatter())
        p.feed(ifile.read())
        l = p.anchorlist
        #soup = BeautifulSoup(ifile.read())
        # l = [str(a['href']) for a in soup.findAll('a')]
    finally:
        ifile.close()
    d = {}
    if url[-1] != '/': # make sure url ends in trailing /
        url += '/'
    for s in l: # find all anchors in the document
        if fileFilter(s): # save it
            o = klass(urllib.basejoin(url, s))
            o.__doc__ = fileDocumenter(s)
            d[fileNamer(s)] = o
    return d


def save_NLMSA_downloaders(url, fileFilter=lambda x: x.endswith(".txt.gz"),
                           resourceStem='Bio.MSA.UCSC.',
                           fileDocumenter=None, fileNamer=None):
    '''save NLMSA downloader / builder objects for a set
    of downloadable textdump files'''
    if fileDocumenter is None:
        fileDocumenter = lambda x: 'NLMSA alignment ' + x
    if fileNamer is None: # a default resource naming function
        fileNamer = lambda x: resourceStem + x[:-3] # remove .gz suffix
    from pygr.nlmsa_utils import NLMSABuilder
    from pygr.downloader import SourceURL
    d = catalog_downloads(url, fileFilter, fileNamer,
                          fileDocumenter, SourceURL)
    for resID, o in d.items():
        nlmsa = NLMSABuilder(o)
        nlmsa.__doc__ = fileDocumenter(resID)
        d[resID[:-4]] = nlmsa # remove .txt suffix
    from pygr import worldbase
    worldbase.add_resource(d)
    worldbase.commit()
    return d # just in case the user wants to see what was saved



import cnestedlist
from nlmsa_utils import EmptySliceError, EmptySlice
import sequence


class NLMSAServer(cnestedlist.NLMSA):
    'serves NLMSA via serializable method calls for XMLRPC'
    xmlrpc_methods = {'getSlice': 0, 'getInfo': 0}

    def getSlice(self, seqID, start, stop):
        'perform an interval query and return results as raw ivals'
        try:
            seq = self.seqDict[seqID]
            nlmsa_id, ns, offset = self.seqs[seq] # GET UNION INFO FOR THIS SEQ
        except KeyError:
            return '' # failure code
        ival = sequence.absoluteSlice(seq, start, stop) # GET THE INTERVAL
        try:
            myslice = self[ival] # DO THE QUERY
        except EmptySliceError:
            return 'EMPTY'
        except KeyError:
            return ''  # FAILURE CODE
        ivals = myslice.rawIvals() # GET RAW INTERVAL DATA
        d = {}
        # Save index info for source seq.
        d[nlmsa_id] = self.seqs.IDdict[str(nlmsa_id)]
        for v in ivals: # SAVE INDEX INFO FOR TARGET SEQS
            id = v[2] # target_id NLMSA_ID
            if not self.seqlist.is_lpo(id):
                # Only non-LPO seqs stored in this index.
                d[id] = self.seqs.IDdict[str(id)]
        # XMLRPC can't handle int dictionaries, use a list.
        l = [(key, val) for key, val in d.items()]
        # List of aligned ivals, list of (nlmsa_id, (seqID, nsID)).
        return nlmsa_id, ivals, l

    def getInfo(self):
        'return list of tuples describing NLMSASequences in this NLMSA'
        l = []
        for ns in self.seqlist:
            l.append((ns.id, ns.is_lpo, ns.length, ns.is_union))
        return l


class NLMSAClient(cnestedlist.NLMSA):
    'client for accessing NLMSAServer via XMLRPC'

    def __init__(self, url=None, name=None, idDictClass=dict, **kwargs):
        cnestedlist.NLMSA.__init__(self, mode='xmlrpc',
                                   idDictClass=idDictClass, **kwargs)
        import coordinator
        self.server = coordinator.get_connection(url, name)
        self.url = url
        self.name = name
        l = self.server.getInfo() # READ NS INFO TABLE
        for nsID, is_lpo, nsLength, is_union in l:
            # is_lpo is automatic below.
            ns = cnestedlist.NLMSASequence(self, None, None, 'onDemand',
                                           is_union, nsLength)
            self.addToSeqlist(ns) # ADD THIS TO THE INDEX

    def close(self):
        pass # required interface, but nothing to do

    def doSlice(self, seq):
        '''getSlice from the server, and create an NLMSASlice object
        from results'''
        result = self.server.getSlice(self.seqs.getSeqID(seq), seq.start,
                                      seq.stop)
        if result == '':
            raise KeyError('this interval is not aligned!')
        elif result == 'EMPTY':
            raise EmptySliceError
        id, l, d = result
        for nlmsaID, (seqID, nsID) in d: # SAVE SEQ INFO TO INDEX
            self.seqs.saveSeq(seqID, nsID, 0, nlmsaID)
        return id, l # HAND BACK THE RAW INTEGER INTERVAL DATA

    def __getitem__(self, k):
        'directly call slice without any ID lookup -- will be done server-side'
        try:
            return cnestedlist.NLMSASlice(self.seqlist[0], k.start, k.stop,
                                          -1, -1, k)
        except EmptySliceError:
            return EmptySlice(k)

    def __getstate__(self):
        return dict(url=self.url, name=self.name, seqDict=self.seqDict)



import os
import sys
from classutil import call_subprocess
import logger

# METHODS FOR AUTOMATIC DOWNLOADING OF RESOURCES


def copy_to_file(f, ifile=None, newpath=None, blocksize=8192000):
    'copy from file obj f to ifile (or create newpath if given)'
    if newpath is not None:
        ifile = file(newpath, 'wb') # binary file
    try:
        while True:
            s = f.read(blocksize)
            if s == '':
                break
            ifile.write(s)
    finally:
        if newpath is not None:
            ifile.close()
        f.close()


def do_gunzip(filepath, newpath=None):
    'gunzip the target using Python gzip module'
    from gzip import GzipFile
    if newpath is None:
        newpath = filepath[:-3]
    f = GzipFile(filepath)
    copy_to_file(f, newpath=newpath)
    return newpath


def run_gunzip(filepath, newpath=None):
    'run gunzip program as a sub process'
    if newpath is None:
        newpath = filepath[:-3]
    ifile = open(newpath, 'w+b')
    try:
        if call_subprocess(['gunzip', '-c', filepath], stdout=ifile):
            raise OSError('gunzip "%s" failed!' % filepath)
    finally:
        ifile.close()
    return newpath


def run_unzip(filepath, newpath=None, singleFile=False, **kwargs):
    '''run unzip program as a sub process,
    save to single file newpath if desired.'''
    if newpath is None:
        newpath = filepath[:-4] # DROP THE .zip SUFFIX
    if singleFile: # concatenate all files into newpath
        ifile = file(newpath, 'wb') # copy as binary file
        try:
            status = call_subprocess(['unzip', '-p', filepath], stdout=ifile)
        finally:
            ifile.close()
    else: # just unzip the package as usual
        status = call_subprocess(['unzip', filepath])
    if status != 0:
        raise OSError('unzip "%s" failed!' % filepath)
    return newpath


def create_dir_if_needed(path):
    'ensure that this directory exists, by creating it if needed'
    import os
    if os.path.isdir(path):
        return # directory exists so nothing to do
    create_dir_if_needed(os.path.dirname(path)) # ensure parent exists
    os.mkdir(path) # create this directory


def create_file_with_path(basepath, filepath):
    'create file in write mode, creating parent directory(s) if needed'
    import os.path
    newpath = os.path.join(basepath, filepath)
    create_dir_if_needed(os.path.dirname(newpath))
    return file(newpath, 'wb') # copy as binary file


def do_unzip(filepath, newpath=None, singleFile=False, **kwargs):
    'extract zip archive, to single file given by newpath if desired'
    # WARNING: zipfile module reads entire file into memory!
    if newpath is None:
        newpath = filepath[:-4]
    from zipfile import ZipFile
    t = ZipFile(filepath, 'r')
    try:
        if singleFile: # extract to a single file
            ifile = file(newpath, 'wb') # copy as binary file
            try:
                for name in t.namelist():
                    ifile.write(t.read(name)) # may run out of memory!!
            finally:
                ifile.close()
        else: # extract a bunch of files as usual
            for name in t.namelist():
                ifile = create_file_with_path(newpath, name)
                ifile.write(t.read(name)) # may run out of memory!!
                ifile.close()
    finally:
        t.close()
    return newpath


def do_untar(filepath, mode='r|', newpath=None, singleFile=False, **kwargs):
    'extract tar archive, to single file given by newpath if desired'
    if newpath is None:
        newpath = filepath + '.out'
    import tarfile
    t = tarfile.open(filepath, mode)
    try:
        if singleFile: # extract to a single file
            ifile = file(newpath, 'wb') # copy as binary file
            try:
                for name in t.getnames():
                    f = t.extractfile(name)
                    copy_to_file(f, ifile)
            finally:
                ifile.close()
        else: # extract a bunch of files as usual
            import os
            t.extractall(os.path.dirname(newpath))
    finally:
        t.close()
    return newpath


def uncompress_file(filepath, **kwargs):
    '''stub for applying appropriate uncompression based on file suffix
    (.tar .tar.gz .tgz .tar.bz2 .gz and .zip for now)'''
    if filepath.endswith('.zip'):
        logger.info('unzipping %s...' % filepath)
        try:
            return run_unzip(filepath, **kwargs)
        except OSError:
            return do_unzip(filepath, **kwargs)
    elif filepath.endswith('.tar'):
        logger.info('untarring %s...' % filepath)
        return do_untar(filepath, newpath=filepath[:-4], **kwargs)
    elif filepath.endswith('.tgz'):
        logger.info('untarring %s...' % filepath)
        return do_untar(filepath, mode='r:gz', newpath=filepath[:-4], **kwargs)
    elif filepath.endswith('.tar.gz'):
        logger.info('untarring %s...' % filepath)
        return do_untar(filepath, mode='r:gz', newpath=filepath[:-7], **kwargs)
    elif filepath.endswith('.tar.bz2'):
        logger.info('untarring %s...' % filepath)
        return do_untar(filepath, mode='r:bz2', newpath=filepath[:-8],
                        **kwargs)
    elif filepath.endswith('.gz'):
        logger.info('gunzipping %s...' % filepath)
        try:  # could use gzip module, but it's two times slower!!
            return run_gunzip(filepath, **kwargs) # run as sub process
        except OSError: # on Windows, have to run as python module
            return do_gunzip(filepath, **kwargs)

    return filepath # DEFAULT: NOT COMPRESSED, SO JUST HAND BACK FILENAME


def download_monitor(bcount, bsize, totalsize):
    'show current download progress'
    if bcount == 0:
        download_monitor.percentage_last_shown = 0.
    bytes = bcount * bsize
    percentage = bytes * 100. / totalsize
    if percentage >= 10. + download_monitor.percentage_last_shown:
        logger.info('downloaded %s bytes (%2.1f%%)...'
                    % (bytes, percentage))
        download_monitor.percentage_last_shown = percentage


def download_unpickler(path, filename, kwargs):
    'try to download the desired file, and uncompress it if need be'
    import os
    import urllib
    import classutil
    if filename is None:
        filename = os.path.basename(path)
    try:
        dl_dir = os.environ['WORLDBASEDOWNLOAD']
    except KeyError:
        dl_dir = classutil.get_env_or_cwd('PYGRDATADOWNLOAD')
    filepath = os.path.join(dl_dir, filename)
    logger.info('Beginning download of %s to %s...' % (path, filepath))
    t = urllib.urlretrieve(path, filepath, download_monitor)
    logger.info('Download done.')
    filepath = uncompress_file(filepath, **kwargs) # UNCOMPRESS IF NEEDED
    # PATH TO WHERE THIS FILE IS NOW STORED
    o = classutil.SourceFileName(filepath)
    o._saveLocalBuild = True # MARK THIS FOR SAVING IN LOCAL PYGR.DATA
    return o

download_unpickler.__safe_for_unpickling__ = 1


class SourceURL(object):
    '''unpickling this object will trigger downloading of the desired path,
    which will be cached to WORLDBASEDOWNLOAD directory if any.
    The value returned from unpickling will simply be the path to the
    downloaded file, as a SourceFileName'''
    _worldbase_no_cache = True # force worldbase to always re-load this class

    def __init__(self, path, filename=None, **kwargs):
        self.path = path
        self.kwargs = kwargs
        self.filename = filename
        if path.startswith('http:'): # make sure we can read this URL
            import httplib
            conn = httplib.HTTPConnection(path.split('/')[2])
            try:
                conn.request('GET', '/'.join([''] + path.split('/')[3:]))
                r1 = conn.getresponse()
                if r1.status != 200:
                    raise OSError('http GET failed: %d %s, %s'
                                  % (r1.status, r1.reason, path))
            finally:
                conn.close()

    def __reduce__(self):
        return (download_unpickler, (self.path, self.filename, self.kwargs))


def generic_build_unpickler(cname, args, kwargs):
    'does nothing but construct the specified klass with the specified args'
    if cname == 'BlastDB':
        from seqdb import BlastDB as klass
    else:
        raise ValueError('''class name not registered for unpickling security.
Add it to pygr.downloader.generic_build_unpickler if needed: ''' + cname)
    o = klass(*args, **kwargs)
    o._saveLocalBuild = True # MARK FOR LOCAL PYGR.DATA SAVE
    return o
generic_build_unpickler.__safe_for_unpickling__ = 1


class GenericBuilder(object):
    'proxy for constructing the desired klass on unpickling'
    _worldbase_no_cache = True # force worldbase to always re-load this class

    def __init__(self, cname, *args, **kwargs):
        self.cname = cname
        self.args = args
        self.kwargs = kwargs

    def __reduce__(self):
        return (generic_build_unpickler, (self.cname, self.args, self.kwargs))




import datetime
import os
import pickle
import re
import sys
import UserDict
from StringIO import StringIO
from mapping import Collection, Mapping, Graph
from classutil import open_shelve, standard_invert, get_bound_subclass, \
     SourceFileName
from coordinator import XMLRPCServerBase


try:
    nonPortableClasses
except NameError: # DEFAULT LIST OF CLASSES NOT PORTABLE TO REMOTE CLIENTS
    nonPortableClasses = [SourceFileName]


class OneTimeDescriptor(object):
    'provides shadow attribute based on schema'

    def __init__(self, attrName, mdb, **kwargs):
        self.attr=attrName
        self.mdb = mdb

    def __get__(self, obj, objtype):
        try:
            resID = obj._persistent_id # GET ITS RESOURCE ID
        except AttributeError:
            raise AttributeError('attempt to access worldbase attr on \
                                 non-worldbase object')
        target = self.mdb.get_schema_attr(resID, self.attr) #get from mdb
        # Save in __dict__ to evade __setattr__.
        obj.__dict__[self.attr] = target
        return target


class ItemDescriptor(object):
    'provides shadow attribute for items in a db, based on schema'

    def __init__(self, attrName, mdb, invert=False, getEdges=False,
                 mapAttr=None, targetAttr=None, uniqueMapping=False, **kwargs):
        self.attr = attrName
        self.mdb = mdb
        self.invert = invert
        self.getEdges = getEdges
        self.mapAttr = mapAttr
        self.targetAttr = targetAttr
        self.uniqueMapping = uniqueMapping

    def get_target(self, obj):
        'return the mapping object for this schema relation'
        try:
            resID = obj.db._persistent_id # GET RESOURCE ID OF DATABASE
        except AttributeError:
            raise AttributeError('attempt to access worldbase attr on \
                                 non-worldbase object')
        targetDict = self.mdb.get_schema_attr(resID, self.attr)
        if self.invert:
            targetDict = ~targetDict
        if self.getEdges:
            targetDict = targetDict.edges
        return targetDict

    def __get__(self, obj, objtype):
        targetDict = self.get_target(obj)
        if self.mapAttr is not None: # USE mapAttr TO GET ID FOR MAPPING obj
            obj_id = getattr(obj, self.mapAttr)
            if obj_id is None: # None MAPS TO None, SO RETURN IMMEDIATELY
                return None # DON'T BOTHER CACHING THIS
            result = targetDict[obj_id] # MAP USING THE SPECIFIED MAPPING
        else:
            result = targetDict[obj] # NOW PERFORM MAPPING IN THAT RESOURCE...
        if self.targetAttr is not None:
            # Get attribute of the result.
            result = getattr(result, self.targetAttr)
        obj.__dict__[self.attr] = result # CACHE IN THE __dict__
        return result


class ItemDescriptorRW(ItemDescriptor):

    def __set__(self, obj, newTarget):
        if not self.uniqueMapping:
            raise WorldbaseSchemaError(
'''You attempted to directly assign to a graph mapping (x.graph = y)!
Instead, treat the graph like a dictionary: x.graph[y] = edgeInfo''')
        targetDict = self.get_target(obj)
        targetDict[obj] = newTarget
        obj.__dict__[self.attr] = newTarget # CACHE IN THE __dict__


class ForwardingDescriptor(object):
    'forward an attribute request to item from another container'

    def __init__(self, targetDB, attr):
        self.targetDB = targetDB # CONTAINER TO GET ITEMS FROM
        self.attr = attr # ATTRIBUTE TO MAP TO

    def __get__(self, obj, objtype):
        target = self.targetDB[obj.id] # GET target FROM CONTAINER
        return getattr(target, self.attr) # GET DESIRED ATTRIBUTE


class SpecialMethodDescriptor(object):
    'enables shadowing of special methods like __invert__'

    def __init__(self, attrName):
        self.attr = attrName

    def __get__(self, obj, objtype):
        try:
            return obj.__dict__[self.attr]
        except KeyError:
            raise AttributeError('%s has no method %s' % (obj, self.attr))


def addSpecialMethod(obj, attr, f):
    '''bind function f as special method attr on obj.
    obj cannot be an builtin or extension class
    (if so, just subclass it)'''
    import new
    m=new.instancemethod(f, obj, obj.__class__)
    try:
        if getattr(obj, attr) == m: # ALREADY BOUND TO f
            return # ALREADY BOUND, NOTHING FURTHER TO DO
    except AttributeError:
        pass
    else:
        raise AttributeError('%s already bound to a different function' % attr)
    setattr(obj, attr, m) # SAVE BOUND METHOD TO __dict__
    # This does forwarding.
    setattr(obj.__class__, attr, SpecialMethodDescriptor(attr))


def getInverseDB(self):
    'default shadow __invert__ method'
    return self.inverseDB # TRIGGER CONSTRUCTION OF THE TARGET RESOURCE


class WorldbaseNotPortableError(ValueError):
    '''indicates that object has a local data dependency and cannot be
    transferred to a remote client'''
    pass


class WorldbaseNotFoundError(KeyError):
    '''unable to find a loadable resource for the requested worldbase
    identifier from WORLDBASEPATH'''
    pass


class WorldbaseMismatchError(ValueError):
    '''_persistent_id attr on object no longer matches its assigned
    worldbase ID?!?'''
    pass


class WorldbaseEmptyError(ValueError):
    "user hasn't queued anything, so trying to save or rollback is an error"
    pass


class WorldbaseReadOnlyError(ValueError):
    'attempt to write data to a read-only resource database'
    pass


class WorldbaseSchemaError(ValueError):
    "attempt to set attribute to an object not in the database bound by schema"
    pass


class WorldbaseNoModuleError(pickle.PickleError):
    'attempt to pickle a class from a non-importable module'
    pass


class PygrPickler(pickle.Pickler):

    def persistent_id(self, obj):
        '''convert objects with _persistent_id to PYGR_ID strings
        during pickling'''
        import types
        try:
            # Check for unpicklable class (i.e. not loaded
            # via a module import).
            if isinstance(obj, types.TypeType) and \
               obj.__module__ == '__main__':
                raise WorldbaseNoModuleError(
'''You cannot pickle a class from __main__!
To make this class (%s) picklable, it must be loaded via a regular
import statement.''' % obj.__name__)
        except AttributeError:
            pass
        try:
            if not isinstance(obj, types.TypeType) and obj is not self.root:
                try:
                    return 'PYGR_ID:%s' % self.sourceIDs[id(obj)]
                except KeyError:
                    if obj._persistent_id is not None:
                        return 'PYGR_ID:%s' % obj._persistent_id
        except AttributeError:
            pass
        for klass in self.badClasses: # CHECK FOR LOCAL DEPENDENCIES
            if isinstance(obj, klass):
                raise WorldbaseNotPortableError(
'''this object has a local data dependency and cannnot be transferred
to a remote client''')
        return None

    def setRoot(self, obj, sourceIDs={}, badClasses=()):
        '''set obj as root of pickling tree: genuinely pickle it
        (not just its id)'''
        self.root = obj
        self.sourceIDs = sourceIDs
        self.badClasses = badClasses


class MetabaseServer(object):
    'simple XMLRPC resource database server'
    xmlrpc_methods = {'getResource': 0, 'registerServer': 0, 'delResource': 0,
                      'getName': 0, 'dir': 0, 'get_version': 0}
    _pygr_data_version = (0, 1, 0)

    def __init__(self, name, readOnly=True, downloadDB=None):
        self.name = name
        self.d = {}
        self.docs = {}
        self.downloadDB = {}
        self.downloadDocs = {}
        if readOnly: # LOCK THE INDEX.  DON'T ACCEPT FOREIGN DATA!!
            # Only allow these methods!
            self.xmlrpc_methods = {'getResource': 0, 'getName': 0, 'dir': 0,
                                   'get_version': 0}
        if downloadDB is not None:
            self.read_download_db(downloadDB)

    def read_download_db(self, filename, location='default'):
        'add the designated resource DB shelve to our downloadable resources'
        d = open_shelve(filename, 'r')
        for k, v in d.items():
            if k.startswith('__doc__.'): # SAVE DOC INFO FOR THIS ID
                self.downloadDocs[k[8:]] = v
            else: # SAVE OBJECT INFO
                self.downloadDB.setdefault(k, {})[location] = v
        d.close()

    def getName(self):
        'return layer name for this server'
        return self.name

    def get_db(self, download):
        if download: # USE SEPARATE DOWNLOAD DATABASE
            return (self.downloadDB, self.downloadDocs)
        else: # USE REGULAR XMLRPC SERVICES DATABASE
            return (self.d, self.docs)

    def getResource(self, id, download=False):
        'return dict of location:pickleData for requested ID'
        db, docs = self.get_db(download)
        try:
            d = db[id] # RETURN DICT OF PICKLED OBJECTS
        except KeyError:
            return '' # EMPTY STRING INDICATES FAILURE
        if id.startswith('SCHEMA.'): # THIS IS A REQUEST FOR SCHEMA INFO
            for location in d: # -schemaEdge DATA NOT SENDABLE BY XMLRPC
                try:
                    del d[location]['-schemaEdge']
                except KeyError:
                    pass
        else: # THIS IS A REGULAR RESOURCE REQUEST
            try: # PASS ITS DOCSTRING AS A SPECIAL ENTRY
                d['__doc__'] = docs[id]['__doc__']
            except KeyError:
                pass
        return d

    def registerServer(self, locationKey, serviceDict):
        '''add services in serviceDict to this server under the
        specified location'''
        n = 0
        for id, (infoDict, pdata) in serviceDict.items():
            self.d.setdefault(id, {})[locationKey] = pdata # SAVE RESOURCE
            if infoDict is not None:
                self.docs[id] = infoDict
            n += 1
        return n  # COUNT OF SUCCESSFULLY REGISTERED SERVICES

    def delResource(self, id, locationKey):
        'delete the specified resource under the specified location'
        try:
            del self.d[id][locationKey]
            if len(self.d[id]) == 0:
                del self.docs[id]
        except KeyError:
            pass
        return ''  # DUMMY RETURN VALUE FOR XMLRPC

    def dir(self, pattern, asDict=False, matchType='p', download=False):
        'return list or dict of resources matching the specified string'
        db, docs = self.get_db(download)
        if matchType == 'r':
            pattern = re.compile(pattern)
        l = []
        for name in db: # FIND ALL ITEMS WITH MATCHING NAME
            if matchType == 'p' and name.startswith(pattern) or \
               matchType == 'r' and pattern.search(name):
                l.append(name)
        if asDict: # RETURN INFO DICT FOR EACH ITEM
            d = {}
            for name in l:
                d[name] = docs.get(name, {})
            return d
        return l

    def get_version(self):
        return self._pygr_data_version


def raise_illegal_save(self, *l):
    raise WorldbaseReadOnlyError(
'''You cannot save data to a remote XMLRPC server.
Give a user-editable resource database as the first entry
in your WORLDBASEPATH!''')


class XMLRPCMetabase(object):
    'client interface to remote XMLRPC resource database'

    def __init__(self, url, mdb, **kwargs):
        from coordinator import get_connection
        self.server = get_connection(url, 'index')
        self.url=url
        self.mdb = mdb
        self.zoneName = self.server.getName()
        self.writeable = False

    def find_resource(self, id, download=False):
        'get pickledata,docstring for this resource ID from server'
        if download: # SPECIFICALLY ASK SERVER FOR DOWNLOADABLE RESOURCES
            d = self.server.getResource(id, download)
        else: # NORMAL MODE TO GET XMLRPC SERVICES
            d = self.server.getResource(id)
        if d == '':
            raise WorldbaseNotFoundError('resource %s not found' % id)
        try:
            docstring = d['__doc__']
            del d['__doc__']
        except KeyError:
            docstring = None
        for location, objdata in d.items(): # return the first resource found
            return objdata, docstring
        raise KeyError('unable to find %s from remote services' % id)

    def registerServer(self, locationKey, serviceDict):
        'forward registration to the server'
        return self.server.registerServer(locationKey, serviceDict)

    def getschema(self, id):
        'return dict of {attr: {args}}'
        d = self.server.getResource('SCHEMA.' + id)
        if d == '': # NO SCHEMA INFORMATION FOUND
            raise KeyError
        for schemaDict in d.values():
            return schemaDict # HAND BACK FIRST SCHEMA WE FIND
        raise KeyError

    def dir(self, pattern, matchType='p', asDict=False, download=False):
        'return list or dict of resources matching the specified string'
        if download:
            return self.server.dir(pattern, asDict, matchType, download)
        else:
            return self.server.dir(pattern, asDict, matchType)

    __setitem__ = raise_illegal_save # RAISE USEFUL EXPLANATORY ERROR MESSAGE
    __delitem__ = raise_illegal_save
    setschema = raise_illegal_save
    delschema = raise_illegal_save


class MySQLMetabase(object):
    '''To create a new resource table, call:
    MySQLMetabase("DBNAME.TABLENAME", mdb, createLayer="LAYERNAME")
    where DBNAME is the name of your database, TABLENAME is the name
    of the table you want to create, and LAYERNAME is the layer name
    you want to assign it'''
    _pygr_data_version = (0, 1, 0)

    def __init__(self, tablename, mdb, createLayer=None, newZone=None,
                 **kwargs):
        from sqlgraph import get_name_cursor, SQLGraph
        self.tablename, self.cursor, self.serverInfo = \
                get_name_cursor(tablename)
        self.mdb = mdb
        self.writeable = True
        self.rootNames = {}
        # Separate table for schema graph.
        schemaTable = self.tablename + '_schema'
        if createLayer is None:
            createLayer = newZone # use the new parameter
        if createLayer is not None: # CREATE DATABASE FROM SCRATCH
            creation_time = datetime.datetime.now()
            self.cursor.execute('drop table if exists %s' % self.tablename)
            self.cursor.execute('create table %s (pygr_id varchar(255) not \
                                null,location varchar(255) not null,docstring \
                                varchar(255),user varchar(255),creation_time \
                                datetime,pickle_size int,security_code bigint,\
                                info_blob text,objdata text not null,\
                                unique(pygr_id,location))' % self.tablename)
            self.cursor.execute('insert into %s (pygr_id,location,\
                                creation_time,objdata) values (%%s,%%s,%%s,\
                                %%s)' % self.tablename,
                                ('PYGRLAYERNAME', createLayer, creation_time,
                                 'a'))
            # Save version stamp.
            self.cursor.execute('insert into %s (pygr_id,location,objdata) \
                                values (%%s,%%s,%%s)' % self.tablename,
                                ('0version', '%d.%d.%d'
                                 % self._pygr_data_version, 'a'))
            self.zoneName = createLayer
            self.cursor.execute('drop table if exists %s' % schemaTable)
            self.cursor.execute('create table %s (source_id varchar(255) not \
                                null,target_id varchar(255),edge_id \
                                varchar(255),unique(source_id,target_id))'
                                % schemaTable)
        else:
            try:
                n = self.cursor.execute('select location from %s where \
                                        pygr_id=%%s' % self.tablename,
                                        ('PYGRLAYERNAME', ))
            except StandardError:
                print >>sys.stderr, '''%s
Database table %s appears to be missing or has no layer name!
To create this table, call
worldbase.MySQLMetabase("%s", createLayer=<LAYERNAME>)
where <LAYERNAME> is the layer name you want to assign it.
%s''' % ('!' * 40, self.tablename, self.tablename, '!' * 40)
                raise
            if n > 0:
                # Get layer name from the db.
                self.zoneName = self.cursor.fetchone()[0]
            if self.cursor.execute('select location from %s where pygr_id=%%s'
                                   % self.tablename, ('0root', )) > 0:
                for row in self.cursor.fetchall():
                    self.rootNames[row[0]] = None
                mdb.save_root_names(self.rootNames)
        self.graph = SQLGraph(schemaTable, self.cursor, attrAlias=
                              dict(source_id='source_id',
                                   target_id='target_id', edge_id='edge_id'),
                              simpleKeys=True, unpack_edge=SchemaEdge(self))

    def save_root_name(self, name):
        self.rootNames[name] = None
        self.cursor.execute('insert into %s (pygr_id,location,objdata) values \
                            (%%s,%%s,%%s)' % self.tablename, ('0root', name,
                                                              'a'))

    def find_resource(self, id, download=False):
        'get construction rule from mysql, and attempt to construct'
        self.cursor.execute('select location,objdata,docstring from %s where \
                            pygr_id=%%s' % self.tablename, (id, ))
        for location, objdata, docstring in self.cursor.fetchall():
            return objdata, docstring # return first resource found
        raise WorldbaseNotFoundError('unable to construct %s from remote \
                                     services')

    def __setitem__(self, id, obj):
        'add an object to this resource database'
        s = dumps(obj) # PICKLE obj AND ITS DEPENDENCIES
        d = get_info_dict(obj, s)
        self.cursor.execute('replace into %s (pygr_id,location,docstring,user,\
                            creation_time,pickle_size,objdata) values (%%s,\
                            %%s,%%s,%%s,%%s,%%s,%%s)' % self.tablename,
                            (id, 'mysql:' + self.tablename, obj.__doc__,
                             d['user'], d['creation_time'], d['pickle_size'],
                             s))
        root = id.split('.')[0]
        if root not in self.rootNames:
            self.save_root_name(root)

    def __delitem__(self, id):
        'delete this resource and its schema rules'
        if self.cursor.execute('delete from %s where pygr_id=%%s'
                               % self.tablename, (id, )) < 1:
            raise WorldbaseNotFoundError('no resource %s in this database'
                                         % id)

    def registerServer(self, locationKey, serviceDict):
        'register the specified services to mysql database'
        n = 0
        for id, (d, pdata) in serviceDict.items():
            n+=self.cursor.execute('replace into %s (pygr_id,location,\
                                   docstring,user,creation_time,pickle_size,\
                                   objdata) values (%%s,%%s,%%s,%%s,%%s,%%s,\
                                   %%s)' % self.tablename,
                                   (id, locationKey, d['__doc__'], d['user'],
                                    d['creation_time'], d['pickle_size'],
                                    pdata))
        return n

    def setschema(self, id, attr, kwargs):
        'save a schema binding for id.attr --> targetID'
        if not attr.startswith('-'): # REAL ATTRIBUTE
            targetID = kwargs['targetID'] # RAISES KeyError IF NOT PRESENT
        kwdata = dumps(kwargs)
        self.cursor.execute('replace into %s (pygr_id,location,objdata) \
                            values (%%s,%%s,%%s)' % self.tablename,
                            ('SCHEMA.' + id, attr, kwdata))

    def delschema(self, id, attr):
        'delete schema binding for id.attr'
        self.cursor.execute('delete from %s where pygr_id=%%s and location=%%s'
                            % self.tablename, ('SCHEMA.' + id, attr))

    def getschema(self, id):
        'return dict of {attr:{args}}'
        d = {}
        self.cursor.execute('select location,objdata from %s where pygr_id=%%s'
                            % self.tablename, ('SCHEMA.' + id, ))
        for attr, objData in self.cursor.fetchall():
            d[attr] = self.mdb.loads(objData)
        return d

    def dir(self, pattern, matchType='p', asDict=False, download=False):
        'return list or dict of resources matching the specified string'

        if matchType == 'r':
            self.cursor.execute('select pygr_id,docstring,user,creation_time,\
                                pickle_size from %s where pygr_id regexp %%s'
                                % self.tablename, (pattern, ))
        elif matchType == 'p':
            self.cursor.execute('select pygr_id,docstring,user,creation_time,\
                                pickle_size from %s where pygr_id like %%s'
                                % self.tablename, (pattern + '%', ))
        else:
            # Exit now to avoid fetching rows with no query executed
            if asDict:
                return {}
            else:
                return []

        d = {}
        for l in self.cursor.fetchall():
            d[l[0]] = dict(__doc__=l[1], user=l[2], creation_time=l[3],
                           pickle_size=l[4])
        if asDict:
            return d
        else:
            return [name for name in d]


class SchemaEdge(object):
    'provides unpack_edge method for schema graph storage'

    def __init__(self, schemaDB):
        self.schemaDB = schemaDB

    def __call__(self, edgeID):
        'get the actual schema object describing this ID'
        return self.schemaDB.getschema(edgeID)['-schemaEdge']


class ResourceDBGraphDescr(object):
    'this property provides graph interface to schema'

    def __get__(self, obj, objtype):
        g = Graph(filename=obj.dbpath + '_schema', mode='cw', writeNow=True,
                  simpleKeys=True, unpack_edge=SchemaEdge(obj))
        obj.graph = g
        return g


class ShelveMetabase(object):
    '''BerkeleyDB-based storage of worldbase resource databases, using
    the python shelve module.  Users will not need to create instances
    of this class themselves, as worldbase automatically creates one for
    each appropriate entry in your WORLDBASEPATH; if the corresponding
    database file does not already exist, it is automatically created
    for you.'''
    _pygr_data_version = (0, 1, 0)
    graph = ResourceDBGraphDescr() # INTERFACE TO SCHEMA GRAPH

    def __init__(self, dbpath, mdb, mode='r', newZone=None, **kwargs):
        import anydbm
        self.dbpath = os.path.join(dbpath, '.pygr_data') # CONSTRUCT FILENAME
        self.mdb = mdb
        self.writeable = True # can write to this storage
        self.zoneName = None
        try: # OPEN DATABASE FOR READING
            self.db = open_shelve(self.dbpath, mode)
            try:
                mdb.save_root_names(self.db['0root'])
            except KeyError:
                pass
            try:
                self.zoneName = self.db['0zoneName']
            except KeyError:
                pass
        except anydbm.error: # CREATE NEW FILE IF NEEDED
            self.db = open_shelve(self.dbpath, 'c')
            self.db['0version'] = self._pygr_data_version # SAVE VERSION STAMP
            self.db['0root'] = {}
            if newZone is not None:
                self.db['0zoneName'] = newZone
                self.zoneName = newZone

    def reopen(self, mode):
        self.db.close()
        self.db = open_shelve(self.dbpath, mode)

    def find_resource(self, resID, download=False):
        'get an item from this resource database'
        objdata = self.db[resID] # RAISES KeyError IF NOT PRESENT
        try:
            return objdata, self.db['__doc__.' + resID]['__doc__']
        except KeyError:
            return objdata, None

    def __setitem__(self, resID, obj):
        'add an object to this resource database'
        s = dumps(obj) # PICKLE obj AND ITS DEPENDENCIES
        self.reopen('w')  # OPEN BRIEFLY IN WRITE MODE
        try:
            self.db[resID] = s # SAVE TO OUR SHELVE FILE
            self.db['__doc__.' + resID] = get_info_dict(obj, s)
            root = resID.split('.')[0] # SEE IF ROOT NAME IS IN THIS SHELVE
            d = self.db.get('0root', {})
            if root not in d:
                d[root] = None # ADD NEW ENTRY
                self.db['0root'] = d # SAVE BACK TO SHELVE
        finally:
            self.reopen('r') # REOPEN READ-ONLY

    def __delitem__(self, resID):
        'delete this item from the database, with a modicum of safety'
        self.reopen('w')  # OPEN BRIEFLY IN WRITE MODE
        try:
            try:
                del self.db[resID] # DELETE THE SPECIFIED RULE
            except KeyError:
                raise WorldbaseNotFoundError('ID %s not found in %s'
                                            % (resID, self.dbpath))
            try:
                del self.db['__doc__.' + resID]
            except KeyError:
                pass
        finally:
            self.reopen('r') # REOPEN READ-ONLY

    def dir(self, pattern, matchType='p', asDict=False, download=False):
        'generate all item IDs matching the specified pattern'
        if matchType == 'r':
            pattern = re.compile(pattern)
        l = []
        for name in self.db:
            if matchType == 'p' and name.startswith(pattern) or \
               matchType == 'r' and pattern.search(name):
                l.append(name)
        if asDict:
            d = {}
            for name in l:
                d[name] = self.db.get('__doc__.' + name, None)
            return d
        return l

    def setschema(self, resID, attr, kwargs):
        'save a schema binding for resID.attr --> targetID'
        if not attr.startswith('-'): # REAL ATTRIBUTE
            targetID = kwargs['targetID'] # RAISES KeyError IF NOT PRESENT
        self.reopen('w')  # OPEN BRIEFLY IN WRITE MODE
        d = self.db.get('SCHEMA.' + resID, {})
        d[attr] = kwargs # SAVE THIS SCHEMA RULE
        self.db['SCHEMA.' + resID] = d # FORCE shelve TO RESAVE BACK
        self.reopen('r')  # REOPEN READ-ONLY

    def getschema(self, resID):
        'return dict of {attr:{args}}'
        return self.db['SCHEMA.' + resID]

    def delschema(self, resID, attr):
        'delete schema binding for resID.attr'
        self.reopen('w')  # OPEN BRIEFLY IN WRITE MODE
        d=self.db['SCHEMA.' + resID]
        del d[attr]
        self.db['SCHEMA.' + resID] = d # FORCE shelve TO RESAVE BACK
        self.reopen('r')  # REOPEN READ-ONLY

    def __del__(self):
        'close the shelve file when finished'
        self.db.close()


def dumps(obj, **kwargs):
    'pickle to string, using persistent ID encoding'
    src = StringIO()
    pickler = PygrPickler(src) # NEED OUR OWN PICKLER, TO USE persistent_id
    # Root of pickle tree: save even if persistent_id.
    pickler.setRoot(obj, **kwargs)
    pickler.dump(obj) # PICKLE IT
    return src.getvalue() # RETURN THE PICKLED FORM AS A STRING


def get_info_dict(obj, pickleString):
    'get dict of standard info about a resource'
    d = dict(creation_time=datetime.datetime.now(),
             pickle_size=len(pickleString), __doc__=obj.__doc__)
    try:
        d['user'] = os.environ['USER']
    except KeyError:
        d['user'] = None
    return d


class MetabaseBase(object):

    def persistent_load(self, persid):
        'check for PYGR_ID:... format and return the requested object'
        if persid.startswith('PYGR_ID:'):
            return self(persid[8:]) # RUN OUR STANDARD RESOURCE REQUEST PROCESS
        else: # UNKNOWN PERSISTENT ID... NOT FROM PYGR!
            raise pickle.UnpicklingError, 'Invalid persistent ID %s' % persid

    def load(self, resID, objdata, docstring):
        'load the pickled data and all its dependencies'
        obj = self.loads(objdata)
        obj.__doc__ = docstring
        if hasattr(obj, '_saveLocalBuild') and obj._saveLocalBuild:
            saver = self.writer.saver # mdb in which to record local copy
            # SAVE AUTO BUILT RESOURCE TO LOCAL PYGR.DATA
            hasPending = saver.has_pending() # any pending transaction?
            saver.add_resource(resID, obj) # add to queue for commit
            obj._saveLocalBuild = False # NO NEED TO SAVE THIS AGAIN
            if hasPending:
                print >>sys.stderr, \
'''Saving new resource %s to local worldbase...
You must use worldbase.commit() to commit!
You are seeing this message because you appear to be in the middle
of a worldbase transaction.  Ordinarily worldbase would automatically commit
this new downloaded resource, but doing so now would also commit your pending
transaction, which you may not be ready to do!''' % resID
            else: # automatically save new resource
                saver.save_pending() # commit it
        else: # NORMAL USAGE
            obj._persistent_id = resID  # MARK WITH ITS PERSISTENT ID
        self.resourceCache[resID] = obj # SAVE TO OUR CACHE
        self.bind_schema(resID, obj) # BIND SHADOW ATTRIBUTES IF ANY
        return obj

    def loads(self, data):
        'unpickle from string, using persistent ID expansion'
        src = StringIO(data)
        unpickler = pickle.Unpickler(src)
        # We provide persistent lookup.
        unpickler.persistent_load = self.persistent_load
        obj = unpickler.load() # ACTUALLY UNPICKLE THE DATA
        return obj

    def __call__(self, resID, debug=None, download=None, *args, **kwargs):
        'get the requested resource ID by searching all databases'
        try:
            return self.resourceCache[resID] # USE OUR CACHED OBJECT
        except KeyError:
            pass
        debug_state = self.debug # SAVE ORIGINAL STATE
        download_state = self.download
        if debug is not None:
            self.debug = debug
        if download is not None: # apply the specified download mode
            self.download = download
        else: # just use our current download mode
            download = self.download
        try: # finally... TO RESTORE debug STATE EVEN IF EXCEPTION OCCURS.
            self.update(debug=self.debug, keepCurrentPath=True) # load if empty
            for objdata, docstr in self.find_resource(resID, download):
                try:
                    obj = self.load(resID, objdata, docstr)
                    break
                except (KeyError, IOError):
                    # Not in this DB; files not accessible...
                    if self.debug: # PASS ON THE ACTUAL ERROR IMMEDIATELY
                        raise
        finally: # RESTORE STATE BEFORE RAISING ANY EXCEPTION
            self.debug = debug_state
            self.download = download_state
        self.resourceCache[resID] = obj # save to our cache
        return obj

    def bind_schema(self, resID, obj):
        'if this resource ID has any schema, bind its attrs to class'
        try:
            schema = self.getschema(resID)
        except KeyError:
            return # NO SCHEMA FOR THIS OBJ, SO NOTHING TO DO
        self.resourceCache.schemaCache[resID] = schema # cache for speed
        for attr, rules in schema.items():
            if not attr.startswith('-'): # only bind real attributes
                self.bind_property(obj, attr, **rules)

    def bind_property(self, obj, attr, itemRule=False, **kwargs):
        'create a descriptor for the attr on the appropriate obj class'
        try: # SEE IF OBJECT TELLS US TO SKIP THIS ATTRIBUTE
            return obj._ignoreShadowAttr[attr] # IF PRESENT, NOTHING TO DO
        except (AttributeError, KeyError):
            pass # PROCEED AS NORMAL
        if itemRule: # SHOULD BIND TO ITEMS FROM obj DATABASE
            # Class used for constructing items.
            targetClass = get_bound_subclass(obj, 'itemClass')
            descr = ItemDescriptor(attr, self, **kwargs)
        else: # SHOULD BIND DIRECTLY TO obj VIA ITS CLASS
            targetClass = get_bound_subclass(obj)
            descr = OneTimeDescriptor(attr, self, **kwargs)
        setattr(targetClass, attr, descr) # BIND descr TO targetClass.attr
        if itemRule:
            try: # BIND TO itemSliceClass TOO, IF IT EXISTS...
                targetClass = get_bound_subclass(obj, 'itemSliceClass')
            except AttributeError:
                pass # NO itemSliceClass, SO SKIP
            else: # BIND TO itemSliceClass
                setattr(targetClass, attr, descr)
        if attr == 'inverseDB': # ADD SHADOW __invert__ TO ACCESS THIS
            addSpecialMethod(obj, '__invert__', getInverseDB)

    def get_schema_attr(self, resID, attr):
        'actually retrieve the desired schema attribute'
        try: # GET SCHEMA FROM CACHE
            schema = self.resourceCache.schemaCache[resID]
        except KeyError:
            # Hmm, it should be cached! Obtain from resource DB and cache.
            schema = self.getschema(resID)
            self.resourceCache.schemaCache[resID] = schema
        try:
            schema = schema[attr] # GET SCHEMA FOR THIS SPECIFIC ATTRIBUTE
        except KeyError:
            raise AttributeError('no worldbase schema info for %s.%s' \
                                 % (resID, attr))
        targetID = schema['targetID'] # GET THE RESOURCE ID
        return self(targetID) # actually load the resource

    def add_root_name(self, name):
        'add name to the root of our data namespace and schema namespace'
        # This forces the root object to add name if not present.
        getattr(self.Data, name)
        getattr(self.Schema, name)

    def save_root_names(self, rootNames):
        'add set of names to our namespace root'
        for name in rootNames:
            self.add_root_name(name)

    def clear_cache(self):
        'clear all resources from cache'
        self.resourceCache.clear()

    def get_writer(self):
        'return writeable mdb if available, or raise exception'
        try:
            return self.writer
        except AttributeError:
            raise WorldbaseReadOnlyError('this metabase is read-only!')

    def add_resource(self, resID, obj=None):
        """assign obj as the specified resource ID to our metabase.
        if obj is None, treat resID as a dictionary whose keys are
        resource IDs and values are the objects to save."""
        if obj is None:
            self.get_writer().saver.add_resource_dict(resID)
        else:
            self.get_writer().saver.add_resource(resID, obj)

    def delete_resource(self, resID):
        'delete specified resource ID from our metabase'
        self.get_writer().saver.delete_resource(resID)

    def commit(self):
        'save any pending resource assignments and schemas'
        self.get_writer().saver.save_pending()

    def rollback(self):
        'discard any pending resource assignments and schemas'
        self.get_writer().saver.rollback()

    def queue_schema_obj(self, schemaPath, attr, schemaObj):
        'add a schema to the list of pending schemas to commit'
        self.get_writer().saver.queue_schema_obj(schemaPath, attr, schemaObj)

    def add_schema(self, resID, schemaObj):
        'assign a schema relation object to a worldbase resource name'
        l = resID.split('.')
        schemaPath = SchemaPath(self, '.'.join(l[:-1]))
        setattr(schemaPath, l[-1], schemaObj)

    def list_pending(self):
        return self.get_writer().saver.list_pending()


class Metabase(MetabaseBase):

    def __init__(self, dbpath, resourceCache, zoneDict=None, parent=None,
                 **kwargs):
        '''zoneDict provides a mechanism for the caller to request information
        about what type of metabase this dbpath mapped to.  zoneDict must
        be a dict'''
        self.parent = parent
        self.Schema = SchemaPath(self)
        self.Data = ResourceRoot(self) # root of namespace
        self.resourceCache = resourceCache
        self.debug = True # single mdb should expose all errors
        self.download = False
        if zoneDict is None: # user doesn't want zoneDict info
            zoneDict = {} # use a dummy dict, disposable
        if dbpath.startswith('http://'):
            storage = XMLRPCMetabase(dbpath, self, **kwargs)
            if 'remote' not in zoneDict:
                zoneDict['remote'] = self
        elif dbpath.startswith('mysql:'):
            storage = MySQLMetabase(dbpath[6:], self, **kwargs)
            if 'MySQL' not in zoneDict:
                zoneDict['MySQL'] = self
        else: # TREAT AS LOCAL FILEPATH
            dbpath = os.path.expanduser(dbpath)
            storage = ShelveMetabase(dbpath, self, **kwargs)
            if dbpath == os.path.expanduser('~') or \
               dbpath.startswith(os.path.expanduser('~') + os.sep):
                if 'my' not in zoneDict:
                    zoneDict['my'] = self
            elif os.path.isabs(dbpath):
                if 'system' not in zoneDict:
                    zoneDict['system'] = self
            elif dbpath.split(os.sep)[0] == os.curdir:
                if 'here' not in zoneDict:
                    zoneDict['here'] = self
            elif 'subdir' not in zoneDict:
                zoneDict['subdir'] = self
        self.storage = storage
        if storage.zoneName is not None and storage.zoneName not in zoneDict:
            zoneDict[storage.zoneName] = self  # record this zone name
        if storage.writeable:
            self.writeable = True
            self.saver = ResourceSaver(self)
            self.writer = self # record downloaded resources here
        else:
            self.writeable = False

    def update(self, worldbasePath=None, debug=None, keepCurrentPath=False):
        if not keepCurrentPath: # metabase has fixed path
            raise ValueError('You cannot change the path of a Metabase')

    def find_resource(self, resID, download=False):
        yield self.storage.find_resource(resID, download)

    def get_pending_or_find(self, resID, **kwargs):
        'find resID even if only pending (not actually saved yet)'
        try: # 1st LOOK IN PENDING QUEUE
            return self.saver.pendingData[resID]
        except KeyError:
            pass
        return self(resID, **kwargs)

    def getschema(self, resID):
        'return dict of {attr: {args}} or KeyError if not found'
        return self.storage.getschema(resID)

    def save_root_names(self, rootNames):
        if self.parent is not None: # add names to parent's namespace as well
            self.parent.save_root_names(rootNames)
        MetabaseBase.save_root_names(self, rootNames) # call the generic method

    def saveSchema(self, resID, attr, args):
        '''save an attribute binding rule to the schema; DO NOT use this
        internal interface unless you know what you are doing!'''
        self.storage.setschema(resID, attr, args)

    def saveSchemaEdge(self, schema):
        'save schema edge to schema graph'
        self.saveSchema(schema.name, '-schemaEdge', schema)
        self.storage.graph += schema.sourceDB # ADD NODE TO SCHEMA GRAPH
        # Edge
        self.storage.graph[schema.sourceDB][schema.targetDB] = schema.name

    def dir(self, pattern='', matchType='p', asDict=False, download=False):
        return self.storage.dir(pattern, matchType, asDict=asDict,
                            download=download)


class ZoneDict(UserDict.DictMixin):
    'interface to current zones'

    def __init__(self, mdbList):
        self.mdbList = mdbList

    def __getitem__(self, zoneName):
        self.mdbList.update(keepCurrentPath=True) # make sure metabases loaded
        return self.mdbList.zoneDict[zoneName]

    def keys(self):
        self.mdbList.update(keepCurrentPath=True) # make sure metabases loaded
        return self.mdbList.zoneDict.keys()

    def copy(self):
        self.mdbList.update(keepCurrentPath=True) # make sure metabases loaded
        return self.mdbList.zoneDict.copy()


class MetabaseList(MetabaseBase):
    '''Primary interface for worldbase resource database access.
    A single instance of this class is created upon import of the
    worldbase module, accessible as worldbase.getResource.  Users
    normally will have no need to create additional instances of this
    class themselves.'''
    # DEFAULT WORLDBASEPATH: HOME, CURRENT DIR, XMLRPC IN THAT ORDER
    defaultPath = ['~', '.', 'http://biodb2.bioinformatics.ucla.edu:5000']

    def __init__(self, worldbasePath=None, resourceCache=None, separator=',',
                 mdbArgs={}):
        '''initializes attrs; does not connect to metabases'''
        if resourceCache is None: # create a cache for loaded resources
            resourceCache = ResourceCache()
        self.resourceCache = resourceCache
        self.mdb = None
        self.mdbArgs = mdbArgs
        self.zoneDict = {}
        self.zones = ZoneDict(self) # interface to dict of zones
        self.worldbasePath = worldbasePath
        self.separator = separator
        self.Schema = SchemaPath(self)
        self.Data = ResourceRoot(self, zones=self.zones) # root of namespace
        self.debug = False # if one load attempt fails, try other metabases
        self.download = False
        self.ready = False

    def get_writer(self):
        'ensure that metabases are loaded, before looking for our writer'
        self.update(keepCurrentPath=True) # make sure metabases loaded
        return MetabaseBase.get_writer(self) # proceed as usual

    def find_resource(self, resID, download=False):
        'search our metabases for pickle string and docstr for resID'
        for mdb in self.mdb:
            try:
                yield mdb.find_resource(resID, download).next()
            except KeyError: # not in this db
                pass
        raise WorldbaseNotFoundError('unable to find %s in WORLDBASEPATH'
                                     % resID)

    def get_worldbase_path(self):
        'get environment var, or default in that order'
        try:
            return os.environ['WORLDBASEPATH']
        except KeyError:
            try:
                return os.environ['PYGRDATAPATH']
            except KeyError:
                return self.separator.join(self.defaultPath)

    def update(self, worldbasePath=None, debug=None, keepCurrentPath=False,
               mdbArgs=None):
        'get the latest list of resource databases'
        if keepCurrentPath: # only update if self.worldbasePath is None
            worldbasePath = self.worldbasePath
        if worldbasePath is None: # get environment var or default
            worldbasePath = self.get_worldbase_path()
        if debug is None:
            debug = self.debug
        if mdbArgs is None:
            mdbArgs = self.mdbArgs
        if not self.ready or self.worldbasePath != worldbasePath: # reload
            self.worldbasePath = worldbasePath
            try: # disconnect from previous writeable interface if any
                del self.writer
            except AttributeError:
                pass
            self.mdb = []
            try: # default: we don't have a writeable mdb to save data in
                del self.writer
            except AttributeError:
                pass
            self.zoneDict = {}
            for dbpath in worldbasePath.split(self.separator):
                try: # connect to metabase
                    mdb = Metabase(dbpath, self.resourceCache, self.zoneDict,
                                   self, **mdbArgs)
                except (KeyboardInterrupt, SystemExit):
                    raise # DON'T TRAP THESE CONDITIONS
                # FORCED TO ADOPT THIS STRUCTURE BECAUSE xmlrpc RAISES
                # socket.gaierror WHICH IS NOT A SUBCLASS OF StandardError...
                # SO I CAN'T JUST TRAP StandardError, UNFORTUNATELY...
                except: # trap errors and continue to next metabase
                    if debug:
                        raise # expose the error immediately
                    else: # warn the user but keep going...
                        import traceback
                        traceback.print_exc(10, sys.stderr)
                        print >>sys.stderr, '''
WARNING: error accessing metabase %s.  Continuing...''' % dbpath
                else: # NO PROBLEM, SO ADD TO OUR RESOURCE DB LIST
                    # Save to our list of resource databases.
                    self.mdb.append(mdb)
                    if mdb.writeable and not hasattr(self, 'writer'):
                        self.writer = mdb # record as place to save resources
            self.ready = True # metabases successfully loaded

    def get_pending_or_find(self, resID, **kwargs):
        'find resID even if only pending (not actually saved yet)'
        for mdb in self.mdb:
            try: # 1st LOOK IN PENDING QUEUE
                return mdb.saver.pendingData[resID]
            except KeyError:
                pass
        return self(resID, **kwargs)

    def registerServer(self, locationKey, serviceDict):
        'register the serviceDict with the first index server in WORLDBASEPATH'
        for mdb in self.mdb:
            if hasattr(mdb.storage, 'registerServer'):
                n = mdb.storage.registerServer(locationKey, serviceDict)
                if n == len(serviceDict):
                    return n
        raise ValueError('unable to register services.  Check WORLDBASEPATH')

    def getschema(self, resID):
        'search our resource databases for schema info for the desired ID'
        for mdb in self.mdb:
            try:
                return mdb.getschema(resID) # TRY TO OBTAIN FROM THIS DATABASE
            except KeyError:
                pass # NOT IN THIS DB
        raise KeyError('no schema info available for ' + resID)

    def dir(self, pattern='', matchType='p', asDict=False, download=False):
        'get list or dict of resources beginning with the specified string'
        self.update(keepCurrentPath=True) # make sure metabases loaded
        results = []
        for mdb in self.mdb:
            results.append(mdb.dir(pattern, matchType, asDict=asDict,
                                   download=download))
        if asDict: # merge result dictionaries
            d = {}
            results.reverse() # give first results highest precedence
            for subdir in results:
                d.update(subdir)
            return d
        else: # simply remove redundancy from results
            d = {}
            for l in results:
                filter(d.setdefault, l) # add all entries to dict
            results = d.keys()
            results.sort()
            return results


class ResourceCache(dict):
    'provide one central repository of loaded resources & schema info'

    def __init__(self):
        dict.__init__(self)
        self.schemaCache = {}

    def clear(self):
        dict.clear(self) # clear our dictionary
        self.schemaCache.clear() #


class ResourceSaver(object):
    'queues new resources until committed to our mdb'

    def __init__(self, mdb):
        self.clear_pending()
        self.mdb = mdb

    def clear_pending(self):
        self.pendingData = {} # CLEAR THE PENDING QUEUE
        self.pendingSchema = {} # CLEAR THE PENDING QUEUE
        self.lastData = {}
        self.lastSchema = {}
        self.rollbackData = {} # CLEAR THE ROLLBACK CACHE

    def check_docstring(self, obj):
        '''enforce requirement for docstring, by raising exception
        if not present'''
        try:
            if obj.__doc__ is None or (hasattr(obj.__class__, '__doc__') and
                                       obj.__doc__==obj.__class__.__doc__):
                raise AttributeError
        except AttributeError:
            raise ValueError('to save a resource object, you MUST give it a \
                             __doc__ string attribute describing it!')

    def add_resource(self, resID, obj):
        'queue the object for saving to our metabase as <resID>'
        self.check_docstring(obj)
        obj._persistent_id = resID # MARK OBJECT WITH ITS PERSISTENT ID
        self.pendingData[resID] = obj # ADD TO QUEUE
        try:
            self.rollbackData[resID] = self.mdb.resourceCache[resID]
        except KeyError:
            pass
        self.cache_if_appropriate(resID, obj)

    def cache_if_appropriate(self, resID, obj):
        try:
            if obj._worldbase_no_cache:
                return # do not cache this object; it is not ready to use!!
        except AttributeError:
            pass
        self.mdb.resourceCache[resID] = obj # SAVE TO OUR CACHE

    def add_resource_dict(self, d):
        'queue a dict of name:object pairs for saving to metabase'
        for k, v in d.items():
            self.add_resource(k, v)

    def queue_schema_obj(self, schemaPath, attr, schemaObj):
        'add a schema object to the queue for saving to our metabase'
        resID = schemaPath.getPath(attr) # GET STRING ID
        self.pendingSchema[resID] = (schemaPath, attr, schemaObj)

    def save_resource(self, resID, obj):
        'save the object as <id>'
        self.check_docstring(obj)
        if obj._persistent_id != resID:
            raise WorldbaseMismatchError(
'''The _persistent_id attribute for %s has changed!
If you changed it, shame on you!  Otherwise, this should not happen, so report
the reproducible steps to this error message as a bug report.''' % resID)
        # Finally, save the object to the database.
        self.mdb.storage[resID] = obj
        self.cache_if_appropriate(resID, obj) # SAVE TO OUR CACHE

    def has_pending(self):
        'return True if there are resources pending to be committed'
        return len(self.pendingData) > 0 or len(self.pendingSchema) > 0

    def save_pending(self):
        'save any pending worldbase resources and schema'
        if len(self.pendingData) > 0 or len(self.pendingSchema) > 0:
            d = self.pendingData
            schemaDict = self.pendingSchema
        else:
            raise WorldbaseEmptyError('there is no data queued for saving!')
        for resID, obj in d.items(): # now save the data
            self.save_resource(resID, obj)
        for schemaPath, attr, schemaObj in schemaDict.values():# save schema
            schemaObj.saveSchema(schemaPath, attr, self.mdb) # save each rule
        self.clear_pending() # FINALLY, CLEAN UP...
        self.lastData = d # keep as a historical record
        self.lastSchema = schemaDict

    def list_pending(self):
        'return tuple of pending data dictionary, pending schema'
        return list(self.pendingData), list(self.pendingSchema)

    def rollback(self):
        'dump any pending data without saving, and restore state of cache'
        if len(self.pendingData) == 0 and len(self.pendingSchema) == 0:
            raise WorldbaseEmptyError('there is no data queued for saving!')
        # Restore the rollback queue.
        self.mdb.resourceCache.update(self.rollbackData)
        self.clear_pending()

    def delete_resource(self, resID): # incorporate this into commit-process?
        'delete the specified resource from resourceCache, saver and schema'
        del self.mdb.storage[resID] # delete from the resource database
        try:
            del self.mdb.resourceCache[resID] # delete from cache if exists
        except KeyError:
            pass
        try:
            del self.pendingData[resID] # delete from queue if exists
        except KeyError:
            pass
        self.delSchema(resID)

    def delSchema(self, resID):
        'delete schema bindings TO and FROM this resource ID'
        storage = self.mdb.storage
        try:
            d = storage.getschema(resID) # GET THE EXISTING SCHEMA
        except KeyError:
            return # no schema stored for this object so nothing to do...
        # This is more aggressive than needed... Could be refined.
        self.mdb.resourceCache.schemaCache.clear()
        for attr, obj in d.items():
            if attr.startswith('-'): # A SCHEMA OBJECT
                obj.delschema(storage) # DELETE ITS SCHEMA RELATIONS
            storage.delschema(resID, attr) # delete attribute schema rule

    def __del__(self):
        try:
            self.save_pending() # SEE WHETHER ANY DATA NEEDS SAVING
            print >>sys.stderr, '''
WARNING: saving worldbase pending data that you forgot to save...
Remember in the future, you must issue the command worldbase.commit() to save
your pending worldbase resources to your resource database(s), or alternatively
worldbase.rollback() to dump those pending data without saving them.
It is a very bad idea to rely on this automatic attempt to save your
forgotten data, because it is possible that the Python interpreter
may never call this function at exit (for details see the atexit module
docs in the Python Library Reference).'''
        except WorldbaseEmptyError:
            pass


class ResourceServer(XMLRPCServerBase):
    'serves resources that can be transmitted on XMLRPC'

    def __init__(self, mdb, name, serverClasses=None, clientHost=None,
                 withIndex=True, excludeClasses=None, downloadDB=None,
                 resourceDict=None, **kwargs):
        'construct server for the designated classes'
        XMLRPCServerBase.__init__(self, name, **kwargs)
        self.mdb = mdb
        if resourceDict is None:
            resourceDict = mdb.resourceCache
        if excludeClasses is None: # DEFAULT: NO POINT IN SERVING SQL TABLES...
            from sqlgraph import SQLTableBase, SQLGraphClustered
            excludeClasses = [SQLTableBase, SQLGraphClustered]
        if serverClasses is None: # DEFAULT TO ALL CLASSES WE KNOW HOW TO SERVE
            from seqdb import SequenceFileDB, BlastDB, \
                 XMLRPCSequenceDB, BlastDBXMLRPC, \
                 AnnotationDB, AnnotationClient, AnnotationServer
            serverClasses=[(SequenceFileDB, XMLRPCSequenceDB, BlastDBXMLRPC),
                           (BlastDB, XMLRPCSequenceDB, BlastDBXMLRPC),
                           (AnnotationDB, AnnotationClient, AnnotationServer)]
            try:
                from cnestedlist import NLMSA
                from xnestedlist import NLMSAClient, NLMSAServer
                serverClasses.append((NLMSA, NLMSAClient, NLMSAServer))
            except ImportError: # cnestedlist NOT INSTALLED, SO SKIP...
                pass
        if clientHost is None: # DEFAULT: USE THE SAME HOST STRING AS SERVER
            clientHost = self.host
        clientDict = {}
        for id, obj in resourceDict.items():
            # Save all objects matching serverClasses.
            skipThis = False
            for skipClass in excludeClasses: # CHECK LIST OF CLASSES TO EXCLUDE
                if isinstance(obj, skipClass):
                    skipThis = True
                    break
            if skipThis:
                continue # DO NOT INCLUDE THIS OBJECT IN SERVER
            skipThis = True
            for baseKlass, clientKlass, serverKlass in serverClasses:
                if isinstance(obj, baseKlass) and not isinstance(obj,
                                                                 clientKlass):
                    skipThis = False # OK, WE CAN SERVE THIS CLASS
                    break
            if skipThis: # HAS NO XMLRPC CLIENT-SERVER CLASS PAIRING
                try: # SAVE IT AS ITSELF
                    self.client_dict_setitem(clientDict, id, obj,
                                             badClasses=nonPortableClasses)
                except WorldbaseNotPortableError:
                    pass # HAS NON-PORTABLE LOCAL DEPENDENCIES, SO SKIP IT
                continue # GO ON TO THE NEXT DATA RESOURCE
            try: # TEST WHETHER obj CAN BE RE-CLASSED TO CLIENT / SERVER
                # Convert to server class for serving.`
                obj.__class__ = serverKlass
            except TypeError: # GRR, EXTENSION CLASS CAN'T BE RE-CLASSED...
                state = obj.__getstate__() # READ obj STATE
                newobj = serverKlass.__new__(serverKlass) # ALLOCATE NEW OBJECT
                newobj.__setstate__(state) # AND INITIALIZE ITS STATE
                obj = newobj # THIS IS OUR RE-CLASSED VERSION OF obj
            try: # USE OBJECT METHOD TO SAVE HOST INFO, IF ANY...
                obj.saveHostInfo(clientHost, self.port, id)
            except AttributeError: # TRY TO SAVE URL AND NAME DIRECTLY ON obj
                obj.url = 'http://%s:%d' % (clientHost, self.port)
                obj.name = id
            obj.__class__ = clientKlass # CONVERT TO CLIENT CLASS FOR PICKLING
            self.client_dict_setitem(clientDict, id, obj)
            obj.__class__ = serverKlass # CONVERT TO SERVER CLASS FOR SERVING
            self[id] = obj # ADD TO XMLRPC SERVER
        self.registrationData = clientDict # SAVE DATA FOR SERVER REGISTRATION
        if withIndex: # SERVE OUR OWN INDEX AS A STATIC, READ-ONLY INDEX
            myIndex = MetabaseServer(name, readOnly=True, # CREATE EMPTY INDEX
                                     downloadDB=downloadDB)
            self['index'] = myIndex # ADD TO OUR XMLRPC SERVER
            # Add our resources to the index.
            self.register('', '', server=myIndex)

    def client_dict_setitem(self, clientDict, k, obj, **kwargs):
        'save pickle and schema for obj into clientDict'
        pickleString = dumps(obj, **kwargs) # PICKLE THE CLIENT OBJECT, SAVE
        clientDict[k] = (get_info_dict(obj, pickleString), pickleString)
        try: # SAVE SCHEMA INFO AS WELL...
            clientDict['SCHEMA.' + k] = (dict(schema_version='1.0'),
                                         self.mdb.getschema(k))
        except KeyError:
            pass # NO SCHEMA FOR THIS OBJ, SO NOTHING TO DO


class ResourcePath(object):
    'simple way to read resource names as python foo.bar.bob expressions'

    def __init__(self, mdb, base=None):
        self.__dict__['_path'] = base # AVOID TRIGGERING setattr!
        self.__dict__['_mdb'] = mdb

    def getPath(self, name):
        if self._path is not None:
            return self._path + '.' + name
        else:
            return name

    def __getattr__(self, name):
        'extend the resource path by one more attribute'
        attr = self._pathClass(self._mdb, self.getPath(name))
        # MUST NOT USE setattr BECAUSE WE OVERRIDE THIS BELOW!
        self.__dict__[name] = attr # CACHE THIS ATTRIBUTE ON THE OBJECT
        return attr

    def __call__(self, *args, **kwargs):
        'construct the requested resource'
        return self._mdb(self._path, *args, **kwargs)

    def __setattr__(self, name, obj):
        'save obj using the specified resource name'
        self._mdb.add_resource(self.getPath(name), obj)

    def __delattr__(self, name):
        self._mdb.delete_resource(self.getPath(name))
        try: # IF ACTUAL ATTRIBUTE EXISTS, JUST DELETE IT
            del self.__dict__[name]
        except KeyError: # TRY TO DELETE RESOURCE FROM THE DATABASE
            pass # NOTHING TO DO

    def __dir__(self, prefix=None, start=None):
        """return list of our attributes from worldbase search"""
        if prefix is None:
            start = len(self._path) + 1 # skip past . separator
            prefix = self._path
        l = self._mdb.dir(prefix)
        d = {}
        for name in l:
            if name.startswith(prefix):
                d[name[start:].split('.')[0]] = None
        return d.keys()

ResourcePath._pathClass = ResourcePath


class ResourceRoot(ResourcePath):
    'provide proxy to public metabase methods'

    def __init__(self, mdb, base=None, zones=None):
        ResourcePath.__init__(self, mdb, base)
        self.__dict__['schema'] = mdb.Schema # AVOID TRIGGERING setattr!
        if zones is not None:
            self.__dict__['zones'] = zones
        for attr in ('dir', 'commit', 'rollback', 'add_resource',
                     'delete_resource', 'clear_cache', 'add_schema',
                     'update', 'list_pending'):
            self.__dict__[attr] = getattr(mdb, attr) # mirror metabase methods

    def __call__(self, resID, *args, **kwargs):
        """Construct the requested resource"""
        return self._mdb(resID, *args, **kwargs)

    def __dir__(self):
        return ResourcePath.__dir__(self, '', 0)


class ResourceZone(object):
    'provide pygr.Data old-style interface to resource zones'

    def __init__(self, mdb, zoneName):
        self._mdbParent = mdb
        self._zoneName = zoneName

    def __getattr__(self, name):
        # Make sure metabases have been loaded.
        self._mdbParent.update(keepCurrentPath=True)
        try:
            mdb = self._mdbParent.zoneDict[self._zoneName] # get our zone
        except KeyError:
            raise ValueError('no zone "%s" available' % self._zoneName)
        if name == 'schema': # get schema root
            return SchemaPath.__getitem__(self, mdb)
        else: # treat as regular worldbase string
            return ResourcePath.__getitem__(self, mdb, name)


class SchemaPath(ResourcePath):
    'save schema information for a resource'

    def __setattr__(self, name, schema):
        try:
            schema.saveSchema # VERIFY THAT THIS LOOKS LIKE A SCHEMA OBJECT
        except AttributeError:
            raise ValueError('not a valid schema object!')
        self._mdb.queue_schema_obj(self, name, schema) # QUEUE IT

    def __delattr__(self, attr):
        raise NotImplementedError('schema deletion is not yet implemented.')

SchemaPath._pathClass = SchemaPath


class DirectRelation(object):
    'bind an attribute to the target'

    def __init__(self, target):
        self.targetID = getID(target)

    def schemaDict(self):
        return dict(targetID=self.targetID)

    def saveSchema(self, source, attr, mdb, **kwargs):
        d = self.schemaDict()
        d.update(kwargs) # ADD USER-SUPPLIED ARGS
        try: # IF kwargs SUPPLIED A TARGET, SAVE ITS ID
            d['targetID'] = getID(d['targetDB'])
            del d['targetDB']
        except KeyError:
            pass
        mdb.saveSchema(getID(source), attr, d)


class ItemRelation(DirectRelation):
    'bind item attribute to the target'

    def schemaDict(self):
        return dict(targetID=self.targetID, itemRule=True)


class ManyToManyRelation(object):
    'a general graph mapping from sourceDB -> targetDB with edge info'
    _relationCode = 'many:many'

    def __init__(self, sourceDB, targetDB, edgeDB=None, bindAttrs=None,
                 sourceNotNone=None, targetNotNone=None):
        self.sourceDB = getID(sourceDB) # CONVERT TO STRING RESOURCE ID
        self.targetDB = getID(targetDB)
        if edgeDB is not None:
            self.edgeDB = getID(edgeDB)
        else:
            self.edgeDB = None
        self.bindAttrs = bindAttrs
        if sourceNotNone is not None:
            self.sourceNotNone = sourceNotNone
        if targetNotNone is not None:
            self.targetNotNone = targetNotNone

    def save_graph_bindings(self, graphDB, attr, mdb):
        '''save standard schema bindings to graphDB attributes
        sourceDB, targetDB, edgeDB'''
        graphDB = graphDB.getPath(attr) # GET STRING ID FOR source
        self.name = graphDB
        mdb.saveSchemaEdge(self) #SAVE THIS RULE
        b = DirectRelation(self.sourceDB) # SAVE sourceDB BINDING
        b.saveSchema(graphDB, 'sourceDB', mdb)
        b = DirectRelation(self.targetDB) # SAVE targetDB BINDING
        b.saveSchema(graphDB, 'targetDB', mdb)
        if self.edgeDB is not None: # SAVE edgeDB BINDING
            b = DirectRelation(self.edgeDB)
            b.saveSchema(graphDB, 'edgeDB', mdb)
        return graphDB

    def saveSchema(self, path, attr, mdb):
        'save schema bindings associated with this rule'
        graphDB = self.save_graph_bindings(path, attr, mdb)
        if self.bindAttrs is not None:
            bindObj = (self.sourceDB, self.targetDB, self.edgeDB)
            bindArgs = [{}, dict(invert=True), dict(getEdges=True)]
            try: # USE CUSTOM INVERSE SCHEMA IF PROVIDED BY TARGET DB
                bindArgs[1] = mdb.get_pending_or_find(graphDB). \
                        _inverse_schema()
            except AttributeError:
                pass
            for i in range(3):
                if len(self.bindAttrs) > i and self.bindAttrs[i] is not None:
                    b = ItemRelation(graphDB) # SAVE ITEM BINDING
                    b.saveSchema(bindObj[i], self.bindAttrs[i],
                                 mdb, **bindArgs[i])

    def delschema(self, resourceDB):
        'delete resource attribute bindings associated with this rule'
        if self.bindAttrs is not None:
            bindObj = (self.sourceDB, self.targetDB, self.edgeDB)
            for i in range(3):
                if len(self.bindAttrs) > i and self.bindAttrs[i] is not None:
                    resourceDB.delschema(bindObj[i], self.bindAttrs[i])


class OneToManyRelation(ManyToManyRelation):
    _relationCode = 'one:many'


class OneToOneRelation(ManyToManyRelation):
    _relationCode = 'one:one'


class ManyToOneRelation(ManyToManyRelation):
    _relationCode = 'many:one'


class InverseRelation(DirectRelation):
    "bind source and target as each other's inverse mappings"
    _relationCode = 'inverse'

    def saveSchema(self, source, attr, mdb, **kwargs):
        'save schema bindings associated with this rule'
        source = source.getPath(attr) # GET STRING ID FOR source
        self.name = source
        mdb.saveSchemaEdge(self) #SAVE THIS RULE
        DirectRelation.saveSchema(self, source, 'inverseDB',
                                  mdb, **kwargs) # source -> target
        b = DirectRelation(source) # CREATE REVERSE MAPPING
        b.saveSchema(self.targetID, 'inverseDB',
                     mdb, **kwargs) # target -> source

    def delschema(self, resourceDB):
        resourceDB.delschema(self.targetID, 'inverseDB')


def getID(obj):
    'get persistent ID of the object or raise AttributeError'
    if isinstance(obj, str): # TREAT ANY STRING AS A RESOURCE ID
        return obj
    elif isinstance(obj, ResourcePath):
        return obj._path # GET RESOURCE ID FROM A ResourcePath
    else:
        try: # GET RESOURCE'S PERSISTENT ID
            return obj._persistent_id
        except AttributeError:
            raise AttributeError('this obj has no persistent ID!')



from __future__ import generators
from seqref import *
from pygr.seqdb import *
import string


def refIntervals(s):
    begin = 0
    gaps = 0

    end = 0
    for end in range(len(s)):
        if (s[end] == '-'):
            if (begin < end):
                yield (begin, end, begin - gaps, end - gaps, s[begin:end])
            begin = end + 1
            gaps += 1
    if end == 0:
        return
    end = end + 1
    if (begin < end):
        yield (begin, end, begin - gaps, end - gaps, s[begin:end])


def reverse_complement(s):
    compl={'a': 't', 'c': 'g', 'g': 'c', 't': 'a', 'u': 'a', 'n': 'n',
           'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'U': 'A', 'N': 'N'}
    return ''.join([compl.get(c, c) for c in s[::-1]])


class MafParser:
    """
    Parses .maf files as defined by the Haussler dataset. The results
    of parsing are available as pathmapping between the sequences
    in the alignment. The sequences themselves are assumed unknown
    and use AnonSequence class.
    """
    options = {}

    def __init__(self, vbase=''):
        self.mAlign = PathMapping()
        self.sequences = {}
        self.vbase = vbase
        self._vid = 0

    def setpar(self, arry):
        """internal function """
        for p in arry:
            (key, value) = p.split('=')
            self.options[key] = value

    def readalign(self, opt, fh):
        """internal function parses alignment record from .maf file"""
##        print "entering readalign:", opt
        edgeInfo = {}
        for p in opt:
            (key, value) = p.split('=')
            edgeInfo[key] = value

        s = fh.readline().split()
##        print s;
        if(len(s) == 7 and s[0] == 's'):
            vseq = self._vseq(len(s[6]))
            self.mAlign += vseq
        while len(s) == 7 and s[0] == 's':
            # Add the sequence name to the dictionary,
            # then add a corresponding node to the mapping.
            if s[1] not in self.sequences:
                self.sequences[s[1]] = AnonSequence(int(s[5]), s[1])
                self.mAlign += self.sequences[s[1]]

            # PROCESS THE KNOWN INTERVALS
            if(s[4] == '-'):
                ns = self.sequences[s[1]][-int(s[2]):-int(s[2]) - int(s[3])]
                self.sequences[s[1]].seqsplice(reverse_complement(
                    s[6].replace('-', '')), ns.start, ns.stop)
            else:
                ns = self.sequences[s[1]][int(s[2]):int(s[2]) + int(s[3])]
                self.sequences[s[1]].seqsplice(s[6].replace('-', ''),
                                               ns.start, ns.stop)

            for inter in refIntervals(s[6]):
                self.mAlign[vseq[inter[0]:inter[1]]][ns[inter[2]:inter[3]]] = \
                        (inter[4])
                self.mAlign[ns[inter[2]:inter[3]]][vseq[inter[0]:inter[1]]] = \
                        (inter[4])

            s = fh.readline().split()

    def parse(self, filehandle):
        """parses the .maf filehandle """
        l = filehandle.readline()
        if l.split()[0] != '##maf':
            return
        else:
            self.setpar(l.split()[1:])

        l=filehandle.readline()
        while l:
            la = l.split()
##            print la
            if(len(la)==0 or la[0]=='#'):
##                print "skipping"
                1
            elif(la[0]=='a'):
##                print "reading alignment"
                self.readalign(la[1:], filehandle)
            else:
##                print "end of records"
                return

            l=filehandle.readline()

    def _vseq(self, slen):
        alen = len(string.letters)
        uid = self.vbase
        cum = self._vid
        while cum / alen > 0:
            uid += string.letters[cum % alen]
            cum /= alen
        uid += string.letters[cum % alen]
        self._vid += 1
        return AnonSequence(slen, uid)

    def _dump(self, alignTab, sequenceTab=None):
        for row in self.mAlign.repr_dict():
            alignTab.write('\t'.join(map(lambda x: str(x), row.values()))
                           + '\n')

        if(sequenceTab):
            for s in self.sequences.values():
                for inter in s.known_int():
                    sequenceTab.write('\t'.join(map(lambda x: str(x),
                                                    inter.values())) + '\n')

        del self.mAlign
        del self.sequences
        self.mAlign = PathMapping()
        self.sequences = {}

    def parseIntoDB(self, filehandle, cursor, alignTab, sequenceTab=None,
                    update=None):
        """parses the .maf filehandle into database using cursors"""
        c = filehandle.tell()
        filehandle.seek(0, 2)
        filesize = filehandle.tell()
        filehandle.seek(c)
        l = filehandle.readline()
        rc = 0
        count = 0
        if l.split()[0] != '##maf':
            return
        else:
            self.setpar(l.split()[1:])

        l=filehandle.readline()
        while l:
            la = l.split()
##            print la
            if(len(la)==0 or la[0]=='#'):
##                print "skipping"
                1
            elif(la[0]=='a'):
##                print "reading alignment"
                count+=1
                self.readalign(la[1:], filehandle)
                self._dump(alignTab, sequenceTab)
                if(update and not count % 1000):
                    cursor.execute(update % (int(filehandle.tell() * 100.
                                                 / filesize)))
            else:
##                print "end of records"
                return
            l=filehandle.readline()
